/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 Robert-Jan van der Waals                    *
 *                             rwaals@anytimenow.com                          *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.console;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.Image;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.lang.reflect.Method;
import java.util.Enumeration;
import java.util.LinkedList;
import java.util.Vector;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JMenuBar;
import javax.swing.JTabbedPane;

import net.sf.dc.console.components.ComponentFactory;
import net.sf.dc.console.components.Layout;
import net.sf.dc.console.da.ImportCDDialog;
import net.sf.dc.console.note.NotePanel;
import net.sf.dc.console.sql.SQLPanel;
import net.sf.dc.core.IconLibrary;
import net.sf.dc.core.Repository;
import net.sf.dc.core.modules.DCModule;
import net.sf.dc.core.modules.DCModules;
import net.sf.dc.core.objects.DCFieldDefinition;
import net.sf.dc.core.objects.Requestor;
import net.sf.dc.core.objects.RequestorCollection;
import net.sf.dc.db.Database;
import net.sf.dc.db.DatabaseConnection;
import net.sf.dc.messages.Messages;
import net.sf.dc.messages.MessagesWriter;
import net.sf.dc.resources.DCLabels;
import net.sf.dc.resources.DCMessages;
import net.sf.dc.wf.WorkFlow;
import net.sf.ext.jflexsett.settings.Settings;
import net.sf.ext.jflexsett.settings.SettingsFile;
import net.sf.ext.jflexsett.settings.ui.SettingsView;
import net.sf.ext.musicorganizer.OrganizeMp3UI;
import net.sf.ext.util.Resources;
import net.sf.ext.util.Utilities;

/**
 * Holder of all the UI panels. The panels are shown as tabs.
 * By using this frame background colors can be set and requests can be past
 * to the panels by using the @see RequestorCollection class.
 * At the top of the frame a menubar is shown. From this menu the module can
 * be selected, the log form and record form can be opened and the MP3
 * organizer can be selected.
 * 
 * NOTE: I have specifically chosen to load all the panels at once. 
 *       Of course one can say this is heavy on the resources, but.. 
 *       There are just a couple of panels. Switching between modules
 *       should be smooth. Constantly needing to rebuild panels and 
 *       other UI stuff would make the application slow
 *
 * @author Robert-Jan van der Waals
 * @since 1.3
 * @version DC0.50
 */
public class MainFrame extends JFrame {
    
    private JMenuBar toolBar = new JMenuBar();
    private Vector vMenuBarTools = new Vector(0);
    
    private static JTabbedPane tabbedPane;

    private static JButton toolMusicOrganizer;

    private static SQLPanel sqlPanel;
    private static NotePanel notePanel;
    private static SettingsView settingsView;
    private static JComboBox comboModule;

    //private static JMenuBar menuBar;
    
    private DCModule activeModule = null;

    private static boolean firstTime = true;

    /**
     * Initializes all the panels and puts them in a tabbed pane. The menubar
     * is build and its actions specified. The background color for all UI
     * components is set
     */
    public MainFrame() {
        super();
        if (firstTime == true) {
            setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE); 
            
            this.getContentPane().setLayout(Layout.getGBL());
            tabbedPane = ComponentFactory.getTabbedPane();
            tabbedPane.setFont(ComponentFactory.getStandardFont());
            
            // Initializes the settings editor
            createSettingsView();

            try {
                setTitle(Resources.getText(DCLabels.lblMainTitle));
                setResizable(false);
                addWindowListener(new WindowAdapter() {
                    public void windowClosing(WindowEvent e) {
                        finish();
                    }
                });

                Image image = IconLibrary._icoMain.getImage();
                setIconImage(image);
                image.flush();

                initializeMenuBarTools();
                
                sqlPanel = new SQLPanel();
                notePanel = new NotePanel();    

                DCModule module = DCModules.getModule(
                                    Settings.getValueAsInt(Repository.settings.stModule));
                comboModule.setSelectedItem(module);
                
                // apply the last active module
                changeModule(false);
                activeModule = module;

                getContentPane().add(toolBar, Layout.getGBC( 0, 0, 1, 1, 1.0, 1.0
                                     ,GridBagConstraints.NORTHWEST, GridBagConstraints.NONE,
                                     new Insets(0, 0, 0, 0), 0, 0));
                getContentPane().add(tabbedPane, Layout.getGBC( 0, 1, 5, 1, 1.0, 20.0
                                     ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
                                     new Insets(0, 0, 0, 0), 0, 0));

                pack();
                setHiddenFields();
            } catch (Error err) {
                Messages.setMessage(err);
            }
            this.setBackgroundColor(
                    Settings.getValueAsColor(Repository.settings.stColor));
            
            this.setRequiredFields();

            this.setResizable(true);
            this.setSize(Settings.getValueAsDimension(Repository.settings.stMainViewSize));

            Dimension location = Utilities.getCenteredWindowLocation(getSize());
            setLocation(location.width, location.height);

            firstTime = false;
        }
    }
    
    /**
     * Initializes the Settings Vies. JFlexSett is used to create the
     * settings panel. 
     */
    private void createSettingsView() {
        settingsView = new SettingsView("Data Crow Settings Panel");
        settingsView.setDisclaimer(
            "              __                       " + '\n' +
            "       <-----/@@\\----->               " + '\n' +
            "      <-< <  \\\\//  > >->             " + '\n' +
            "        <-<-\\ __ /->->                " + '\n' +
            "        Data /  \\ Crow                " + '\n' +
            "            ^    ^                     " + '\n' +
            "Created By Robert-Jan van der Waals" + '\n' +
            "Original product: CD Library 3.03" + '\n' +
            "" + '\n' +
            "This product includes Hypersonic SQL."  + '\n' +
            "Originally developed by Thomas Mueller and the Hypersonic SQL Group"  + '\n' +
            ""  + '\n' +
            "This product used JFlexSett 0.1 alpha" + '\n' +
            "Created By Robert-Jan van der Waals" + '\n' +
            "" + '\n' +
            "Find me at http://sourceforge.net/projects/datacrow");
        settingsView.setSize(Settings.getValueAsDimension(Repository.settings.stSettingsViewSize));
        Image image = IconLibrary._icoSettings.getImage();
        settingsView.setIconImage(image);
        image.flush();
    }
    
    /**
     * Removes the cache from the cache directory. The cache directory
     * itself will always remain after its creation.
     * 
     * Cache consists mostly out of pictures retrieved from the database
     */
    private void clearCache() {
        try {
            String sCacheDir = Utilities.getCacheDirectory();
            File file = new File(sCacheDir);
            if (file.exists() && file.isDirectory()) {
                String[] files  = file.list();
                
                if (files != null) {
                    for (int i = 0; i < files.length; i++) {
                        File cacheFl = new File (sCacheDir + files[i]);
                        cacheFl.delete();
                    }
                }
            }
        } catch (Exception exp) {
            Messages.setMessage(Resources.getText(DCMessages.msgCouldNotClearCache),
                                Messages._ERROR);
        	Messages.setMessage(exp);
        }
    }

    private void updateMenuBar() {
        toolBar.removeAll();
        toolBar.setBorder(null);
        for (Enumeration enum = vMenuBarTools.elements(); enum.hasMoreElements(); ) {
            toolBar.add((JComponent) enum.nextElement());
        }
        
        JButton[] menuItems = activeModule.getMenuItems();
        for (int i = 0; i < menuItems.length; i++) {
            menuItems[i].setBorder(BorderFactory.createLineBorder(Color.LIGHT_GRAY));
            toolBar.add(menuItems[i]);
        }
        toolBar.add(comboModule);
    }
    
    
    /**
     * Creates the menu
     */
    private void initializeMenuBarTools() {
        JButton toolLogForm =  ComponentFactory.getMenuButton(
                                IconLibrary._icoLog,
                                Resources.getText(DCLabels.lblOpenLog));
        JButton toolSettings = ComponentFactory.getMenuButton(
                                IconLibrary._icoSettings,
                                Resources.getText(DCLabels.lblSettings));
        toolMusicOrganizer =   ComponentFactory.getMenuButton(
                                IconLibrary._icoMp3Organizer,
                                Resources.getText(DCLabels.lblMusicOrganizer));
        JButton toolImportCD = ComponentFactory.getMenuButton(
                                IconLibrary._icoImport,
                                Resources.getText(DCLabels.lblFileImport));

        toolLogForm.setMnemonic(KeyEvent.VK_L);
        toolSettings.setMnemonic(KeyEvent.VK_E);
        toolMusicOrganizer.setMnemonic(KeyEvent.VK_M);
        toolImportCD.setMnemonic(KeyEvent.VK_I);

        toolLogForm.setBorder(BorderFactory.createLineBorder(Color.LIGHT_GRAY));
        toolImportCD.setBorder(BorderFactory.createLineBorder(Color.LIGHT_GRAY));
        toolSettings.setBorder(BorderFactory.createLineBorder(Color.LIGHT_GRAY));
        toolMusicOrganizer.setBorder(BorderFactory.createLineBorder(Color.LIGHT_GRAY));
        
        toolLogForm.addActionListener(new LogFormMenuAction());
        toolMusicOrganizer.addActionListener(new MP3OrganizerFormMenuAction());
        toolSettings.addActionListener(new SettingsFormMenuAction());
        toolImportCD.addActionListener(new ImportCDAction());
        toolImportCD.setSize(toolImportCD.getWidth(), toolSettings.getHeight());
        
        vMenuBarTools.addElement(toolLogForm);
        vMenuBarTools.addElement(toolSettings);
        vMenuBarTools.addElement(toolMusicOrganizer);
        vMenuBarTools.addElement(toolImportCD);

        comboModule = ComponentFactory.getComboBox();
        comboModule.setFont(ComponentFactory.getBoldFont());
        comboModule.setBorder(BorderFactory.createEtchedBorder());
        
        for (Enumeration enum = DCModules.getModules(); enum.hasMoreElements(); ) {
            DCModule module = (DCModule) enum.nextElement();
            if (module.isStandalone()) {
                comboModule.addItem(module);
            }
        }
        
        comboModule.addActionListener(new ModuleAction());
        
        vMenuBarTools.addElement(comboModule);
    }

    /**
     * Applies the required fields for all modules. A required
     * field will mostly be shown as a red header in the table.
     */
    public void setRequiredFields() {
        for (Enumeration enum = DCModules.getModules(); enum.hasMoreElements(); ) {
            DCModule module = (DCModule) enum.nextElement();
            if (module.getSearchPanel() != null) {
                module.getSearchPanel().setRequiredFields(module.getDataCrowObject());
                module.getInsertPanel().setRequiredFields(module.getDataCrowObject());
            }
        }
        this.repaint();
    }
    
    /**
     * Removes all the fields from the table which are not selected as 
     * visible fields. These fields are solely removed from the table,
     * not from the tablemodel. They will be updated and used in validations.
     */
    public static void setHiddenFields() {
        for (Enumeration enum = DCModules.getModules(); enum.hasMoreElements(); ) {
            DCModule module = (DCModule) enum.nextElement();
            int[] hiddenFields = Settings.getValueAsIntegerArray(
                                    module.getHiddenFieldsSettingsKey());
            if (module.getSearchPanel() != null) {
                module.getSearchPanel().panelSearchConds.removeHiddenFields(hiddenFields);
                module.getSearchPanel().removeHiddenFields(hiddenFields);
                module.getInsertPanel().removeHiddenFields(hiddenFields);
            }
            
            DCFieldDefinition[] fields = module.getFieldDefinitions();
            for (int j = 0; j < fields.length; j++) {
                fields[j].setHidden(false);
            }
            
            for (int i = 0; i < hiddenFields.length; i++) {
                fields[hiddenFields[i]].setHidden(true); 
            }
        }
    }

    /**
     * Applies the background color to all panels and their components
     *
     * @color   background color
     */
    public void setBackgroundColor(Color color) {
        for (Enumeration enum = DCModules.getModules(); enum.hasMoreElements(); ) {
            DCModule module = (DCModule) enum.nextElement();
            
            if (module.getSearchPanel() != null) {
                module.getSearchPanel().setBackground(color);
                module.getInsertPanel().setBackground(color);
            }            
        }            
  
        settingsView.setBackgroundColor(color);            
        sqlPanel.setBackground(color);
        notePanel.setBackground(color);
    }

    /**
     * Sets the module. Only the tabs for this module are shown. The module
     * is selected from the module combobox from the menubar
     * 
     * @param compare forces a check if the selected module is the same 
     *                as the current active module. This is basically a hack
     *                for initialization.
     */
    private void changeModule(boolean compare) {
        DCModule module = (DCModule) comboModule.getSelectedItem();
        
        if (compare && module.equals(activeModule)) {
            return;
        } 
        
        activeModule = module;
        
        updateMenuBar();
        
        tabbedPane.removeAll();
        tabbedPane.addTab(Resources.getText(DCLabels.lblExistingValues), 
                          IconLibrary._icoExisting, 
                          module.getSearchPanel());
        tabbedPane.addTab(Resources.getText(DCLabels.lblNewValues), 
                          IconLibrary._icoNew, 
                          module.getInsertPanel());
        tabbedPane.addTab(Resources.getText(DCLabels.lblTabSQL), 
                          IconLibrary._icoSqlManager, 
                          sqlPanel);
        tabbedPane.addTab(Resources.getText(DCLabels.lblTabNote), 
                          IconLibrary._icoNote, 
                          notePanel);
        Settings.setValue(Repository.settings.stModule, 
                          new Integer(module.getIndex()));
    }

    /**
     * Passes requests to the correct module (read, the UI component of
     * the module). 
     * 
     * @param requestors a collection of requests (actions) to be passed 
     *                   to the target
     * @param data data which should be passed to the target. Can be null
     * @param qryWasSuccess Indicates if the query was executed succesfully.
     *                      Unsuccessfull queries can still result in the
     *                      actions being executed (if specified)
     */
    public void displayResults( RequestorCollection requestors, 
                                LinkedList data, 
                                boolean qryWasSuccess) {
        Requestor[] reqObjects = requestors.toRequestorArray();

        for (int i = 0; i < reqObjects.length; i++) {
            Class clazz = reqObjects[i].getClazz();
            String sMethod = reqObjects[i].getMethod();
            Object[] parameters = reqObjects[i].getParameters();

            if (qryWasSuccess || (!qryWasSuccess && reqObjects[i].getExecuteOnFail())) {
                // Specific actions to be performed by the record form
                if (clazz.getName().equals(RecordForm.class.getName())) {
                    if (sMethod.equals("setData")) {
                        RecordForm form = (RecordForm) parameters[0];
                        Object[] row = (Object[]) parameters[1];
                        boolean overwrite = ((Boolean) parameters[2]).booleanValue();
                        int[] destinationFields = (int[]) parameters[3];
                        form.setData(row, overwrite, destinationFields);
                    } 
                } else if (clazz.getName().equals(TablePanel.class.getName())) {
                // Specific actions to be performed by the table panel
                    if (sMethod.equals("closeRecord")) {
                        ((RecordForm) parameters[0]).dispose();
                        parameters[0] = null;
                    }
                } else {
                // Actions to be performed by one of the search / insert / note or
                // sql panels of one of the modules.
                    if (parameters == null && data != null) {
                        parameters = new Object[1];
                        parameters[0] = data;
                    }
                    executeMethod(clazz, sMethod, parameters);
                }    
            }        
        }
    } 
    
    /**
     * Method invocation. Executes the method on the class using the specified
     * parameters. 
     * 
     * @param clazz The class on which the action will be executed
     * @param sMethod The action to be executed
     * @param parameters The parameters for the to be executed action
     */
    private void executeMethod(Class clazz, String sMethod, Object[] parameters) {
        try {
            for (Enumeration enum = DCModules.getModules(); enum.hasMoreElements(); ) {
                 DCModule module = (DCModule) enum.nextElement();
                 
                Object target = null;
                if (clazz.getName().equals(module.getSearchPanel().getClass().getName())) {
                    target = module.getSearchPanel();
                } else if (clazz.getName().equals(module.getInsertPanel().getClass().getName())) {
                    target = module.getInsertPanel();
                } else if (clazz.getName().equals(sqlPanel.getClass().getName())) {
                    target = sqlPanel;
                } else if (clazz.getName().equals(notePanel.getClass().getName())) {
                    target = notePanel;
                }
                
                if (target != null) {
                    Method method = null;
                    
                    boolean found = false;
                    while (clazz != null && method == null) {
                        Method[] methods = clazz.getMethods();
                        for (int i = 0; i < methods.length; i++) {
                            if (methods[i].getName().equals(sMethod)) {
                                method = methods[i];
                                methods = null;
                                clazz = null;
                                found = true;
                                break;
                            }
                        }  
                        if (found) {
                            break;   
                        }
                        methods = null;
                        clazz = null;
                        clazz = clazz.getSuperclass();
                    }
                    method.invoke(target, parameters);
                    return;
                }
            }
        } catch (Exception exp) {
            Messages.setMessage(exp);
        }
    }
    
    /**
     * Closes the application. If running in client / server mode, 
     * a question will popup to close the server as well.
     */
    private void finish() {
        Messages.setMessage(
            Resources.getText(DCMessages.msgApplicationStops),
            Messages._STANDARD);

        if (DatabaseConnection.bServerConnection && 
            Settings.getValueAsBoolean(Repository.settings.stCloseServerQuestion)) {
        
            new QuestionBox(Resources.getText(DCMessages.msgCloseServer));
                               
            if (QuestionBox.result.booleanValue()) {
                WorkFlow wf = new WorkFlow();
                wf.runQueryDirect("SHUTDOWN");
            }
        } else {
            Database.close();
        }

        Settings.setValue(Repository.settings.stMainViewSize, getSize());
        SettingsFile.save();
        clearCache();
        System.exit(0);
    }
    
    private class MP3OrganizerFormMenuAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            OrganizeMp3UI organizeMP3UI = new OrganizeMp3UI();
            organizeMP3UI.setVisible(true);
        }
    }

    private class SettingsFormMenuAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            settingsView.setVisible(true);
        }
    }

    private class LogFormMenuAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            MessagesWriter._windowShow = true;
            MessagesWriter._windowToFront = true;
        }
    }

    private class ImportCDAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            int iModule = Settings.getValueAsInt(Repository.settings.stModule);
            new ImportCDDialog(DCModules.getModule(iModule));
        }
    }

    private class ModuleAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            changeModule(true);
        }
    }
}