/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 Robert-Jan van der Waals                    *
 *                             rwaals@anytimenow.com                          *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.console;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.Image;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.ScrollPaneConstants;

import net.sf.dc.console.components.ComponentFactory;
import net.sf.dc.console.components.Layout;
import net.sf.dc.console.components.LongTextField;
import net.sf.dc.console.components.PictureField;
import net.sf.dc.console.components.ShortTextField;
import net.sf.dc.core.IconLibrary;
import net.sf.dc.core.Repository;
import net.sf.dc.core.modules.DCModule;
import net.sf.dc.core.modules.DCModules;
import net.sf.dc.core.objects.DCObject;
import net.sf.dc.core.objects.Requestor;
import net.sf.dc.core.objects.RequestorCollection;
import net.sf.dc.core.objects.movie.Movie;
import net.sf.dc.core.objects.software.Software;
import net.sf.dc.messages.Messages;
import net.sf.dc.processes.AmazonUpdate;
import net.sf.dc.processes.ImdbUpdate;
import net.sf.dc.resources.DCLabels;
import net.sf.dc.resources.DCMessages;
import net.sf.dc.resources.DCTooltips;
import net.sf.ext.jflexsett.settings.Settings;
import net.sf.ext.util.Resources;
import net.sf.ext.util.StringUtils;
import net.sf.ext.util.SwingUtil;
import net.sf.ext.util.Utilities;


/**
 * Form used to represent one record in the database or to create
 * a new record. The record form is a very generic form and is build up by 
 * using the information present in the module.
 *
 * @author Robert-Jan van der Waals
 * @since 1.3
 * @version DC0.50
 */
public class RecordForm extends JFrame {
    
    private DCObject o;

    private JLabel[] cLabels;
    private JComponent[] cInputFields;
    private JScrollPane[] cScrollPanes;

    private JPanel panelInput = new JPanel(false);
    private JPanel panelPictures = new JPanel(false);
    private JPanel panelAction = new JPanel(false);
    
    private Color color = Settings.getValueAsColor(
                                   Repository.settings.stColor);

    private Class clazz;
    private final boolean update;
    private final int row;

    /**
     * Initializes and shows the record form 
     * 
     * @param updateMode Update / Insert mode
     * @param o Data Crow Object to be displayed
     * @param imdb use imdb update facilities
     * @param amazon use amazon update facilitie
     * @param clazz the clazz which triggered this record to be opened
     * @param row the UI row number
     */
    public RecordForm(  boolean updateMode,
                        DCObject o, 
                        boolean imdb,
                        boolean amazon,
                        Class clazz, 
                        int row) {
                  
        update = updateMode;                            
        SwingUtil.setRootFrame(this);
        this.setResizable(true);
        
        Image img = IconLibrary._icoMain.getImage();
        this.setIconImage(img);
        this.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
        img.flush();
        
        this.o = o;
        this.clazz = clazz;
        this.row = row;
        
        String id = (String) o.getValue(DCObject._ID);
        
        if (!update) {
            setTitle(Resources.getText(DCLabels.lblNewRecord));
        } else {
            setTitle(Resources.getText(DCLabels.lblRecord, id));
        }
        
        addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                close();
            }
        });            
        addMouseListener(new ActiveFrame());
        
        Color colorRequired = ComponentFactory.getRequiredColor();
        String[] labelText = o.getLabels();
        int[] maxFieldLength = o.getMaxFieldLengths();
        int[] componentType = o.getFieldTypes();
        
        cLabels = new JLabel[componentType.length];
        cInputFields = new JComponent[componentType.length];
        cScrollPanes = new JScrollPane[componentType.length];
        
        this.getContentPane().setBackground(color);
        this.setBackground(color);

        // Main panel
        this.getContentPane().setLayout(Layout.getGBL());

        panelInput.setLayout(Layout.getGBL());
        panelPictures.setLayout(Layout.getGBL());
        panelAction.setLayout(Layout.getGBL());

        cLabels[0] = ComponentFactory.getLabel(labelText[0]);
        ShortTextField textField = ComponentFactory.getShortTextField(100);
        textField.setBackground(ComponentFactory.getDisabledColor());
        textField.setEditable(false);

        cInputFields[0] = textField;

        for (int i = 1; i < componentType.length; i ++) {
            cLabels[i] = ComponentFactory.getLabel(labelText[i]);
            JComponent comp = ComponentFactory.getComponent(componentType[i], 
                                                            labelText[i],
                                                            maxFieldLength[i]);
            cInputFields[i]= comp;
        }

        // fill the input panel with the generated fields
        int y = 0;
        int x = 0;
        
        int xPic = 0;
        int yPic = 0;
        
        panelInput.setBackground(color);
        panelPictures.setBackground(color);
        boolean pics = false;
        
        for (int j = 0 ; j < componentType.length; j++) {
            if (!o.isHidden(j)) {
                JComponent comp = cInputFields[j];
                cLabels[j].setBackground(color);
                
                int stretch = GridBagConstraints.HORIZONTAL;
                int factor = 1;
                
                if (componentType[j] == ComponentFactory._LONGTEXTFIELD) {
                    stretch = GridBagConstraints.BOTH; 
                    factor  = 50;
    
                    LongTextField longText = (LongTextField) comp;
                    longText.setMargin(new Insets(0, 0, 0, 0));

                    JScrollPane pane = new JScrollPane(longText);
                    pane.setBorder(ComponentFactory.getBorder());
                    pane.setPreferredSize(new Dimension(100,100));                    
                    pane.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
                    pane.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED);
                    comp = pane;
                    cScrollPanes[j] = pane;
                }
                
                if (!(comp instanceof PictureField)) {
                    panelInput.add(cLabels[j], Layout.getGBC(x, y, 1, 1, 1.0, 1.0
                                              ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                                               new Insets(0, 0, 0, 5), 0, 0));
                    panelInput.add(comp,       Layout.getGBC(x + 1, y, 1, 1, factor, factor
                                              ,GridBagConstraints.NORTHWEST, stretch,
                                               new Insets(0, 0, 0, 0), 0, 0));
                } else { 
                    // picture field
                    pics = true;
                    if (xPic == 3) {
                        xPic = 0;
                        yPic = 2;
                    }
                    panelPictures.add(cLabels[j], Layout.getGBC(xPic, yPic, 1, 1, 1.0, 1.0
                                     ,GridBagConstraints.SOUTHWEST, GridBagConstraints.HORIZONTAL,
                                      new Insets(0, 5, 0, 0), 0, 0));

                    PictureField picField = (PictureField) comp;
                    comp.setPreferredSize(picField.getMinimumSize());
                    panelPictures.add(comp, Layout.getGBC(xPic, yPic + 1, 1, 1, 5.0, 5.0
                                     ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
                                      new Insets(0, 0, 0, 0), 0, 0));
                    xPic++;
                }                
                y++;
            }
        }
        
        int[] required = o.getRequiredFields(); 
        for (int i = 0; i < o.getRequiredFields().length; i++) {
            cLabels[required[i]].setForeground(colorRequired);
        }

        // Fill the actions panel
        JButton buttonClose = ComponentFactory.getButton(Resources.getText(DCLabels.lblClose));
        JButton buttonSave = ComponentFactory.getButton(Resources.getText(DCLabels.lblSave));
        JButton buttonDelete = ComponentFactory.getButton(Resources.getText(DCLabels.lblDelete));
        JButton buttonInternet = null;

        buttonSave.setMnemonic(KeyEvent.VK_S);
        buttonClose.setMnemonic(KeyEvent.VK_C);
        buttonDelete.setMnemonic(KeyEvent.VK_D);

        buttonSave.setToolTipText(Resources.getText(DCTooltips.tpSave));
        buttonDelete.setToolTipText(Resources.getText(DCTooltips.tpDelete));
        buttonClose.setToolTipText(Resources.getText(DCTooltips.tpClose));

        buttonDelete.addActionListener(new DeleteAction());
        buttonClose.addActionListener(new ExitAction());
        buttonSave.addActionListener(new SaveAction());
        
        if (imdb) {
            buttonInternet = ComponentFactory.getButton(Resources.getText(DCLabels.lblIMDBUpdate));
            buttonInternet.addActionListener(new IMDBAction());
            buttonInternet.setToolTipText(Resources.getText(DCTooltips.tpIMDBUpdate));
            buttonInternet.setMnemonic(KeyEvent.VK_U);
        } else if (amazon) {
            buttonInternet = ComponentFactory.getButton(
                             Resources.getText(DCLabels.lblAmazonUpdate));
            buttonInternet.addActionListener(new AmazonAction());
            buttonInternet.setToolTipText(Resources.getText(DCTooltips.tpAmazonUpdate));
            buttonInternet.setMnemonic(KeyEvent.VK_U);
            
        }
        
        panelAction.setBackground(color);
        panelAction.add(buttonSave,         Layout.getGBC(0, 0, 1, 1, 1.0, 1.0
                       ,GridBagConstraints.SOUTHWEST, GridBagConstraints.NONE,
                        new Insets(5, 5, 5, 5), 0, 0));
                        
        if (amazon || imdb) {
            panelAction.add(buttonInternet,   Layout.getGBC(1, 0, 1, 1, 1.0, 1.0
                           ,GridBagConstraints.SOUTHWEST, GridBagConstraints.NONE,
                            new Insets(5, 5, 5, 5), 0, 0));
        }
        
        if (update) {
            panelAction.add(buttonDelete,       Layout.getGBC(2, 0, 1, 1, 1.0, 1.0
                           ,GridBagConstraints.SOUTHWEST, GridBagConstraints.NONE,
                            new Insets(5, 5, 5, 5), 0, 0));
        }
        panelAction.add(buttonClose , Layout.getGBC(3, 0, 1, 1, 1.0, 1.0
                            ,GridBagConstraints.SOUTHWEST, GridBagConstraints.NONE,
                             new Insets(5, 5, 5, 5), 0, 0));


        this.getContentPane().add(    panelInput, Layout.getGBC(0, 0, 1, 1, 1.0, 1.0
                            ,GridBagConstraints.NORTHEAST, GridBagConstraints.BOTH,
                             new Insets(5, 5, 5, 5), 0, 0));
        int width = 1;
        int actionsPos = 1;
        if (pics) {
            width = 2;
            this.getContentPane().add(panelPictures, Layout.getGBC(0, 1, width, 1, 1.0, 1.0
                                     ,GridBagConstraints.NORTHEAST, GridBagConstraints.BOTH,
                                      new Insets(5, 5, 5, 5), 0, 0));
            actionsPos = 2;                       
        }

        this.getContentPane().add(panelAction, Layout.getGBC(0, actionsPos, width, 1, 0.0, 0.0
                      ,GridBagConstraints.SOUTHWEST, GridBagConstraints.NONE,
                       new Insets(5, 5, 5, 5), 0, 0));

        pack();
        
        
        setSize(Settings.getValueAsDimension(getRecordSizeSetting()));
            
        setData(o.getValues(), true, o.getFieldIndices());
        
        Dimension location = Utilities.getCenteredWindowLocation(getSize());
        setLocation(location.width, location.height);
        
        setVisible(true);
    }
    
    /**
     * Retrieves the size settings key for the current active module
     */
    private String getRecordSizeSetting() {
        String setting = Repository.settings.stRecordFormViewSize;
        DCModule module = o.getModule();
        if (module.getIndex() == DCModules._MOVIE) {
            setting = Repository.settings.stMovieRecordFormViewSize;
        } else if (module.getIndex() == DCModules._SOFTWARE) {
            setting = Repository.settings.stSoftwareRecordFormViewSize;
        } else if (module.getIndex() == DCModules._SOFTWARECAT) {
            setting = Repository.settings.stSoftwareCatRecordFormViewSize;
        } else if (module.getIndex() == DCModules._MUSIC) {
            setting = Repository.settings.stMusicRecordFormViewSize;
        } 
        return setting;
    }
    
    
    /**
     * Applies data to the record form
     *
     * @param values    values to be displayed. Values is based on all fields, both
     *                  hidden and visible
     * @param overwrite indicates if existing values should be overwritten
     * @param destinationFields the field indices which are targeted to be filled
     *                  with the values
     */
    public void setData(Object[] values, 
                        boolean overwrite, 
                        int[] destinationFields) {
        for (int i = 0; i < destinationFields.length; i++) {
            Object obj = ComponentFactory.getComponentValue(
                    o.getFieldType(destinationFields[i]),
                    cInputFields[destinationFields[i]]);
            
            boolean empty = obj == null;
            
            if (!empty) {
                empty = StringUtils.isEmpty(obj.toString());
            }
            
            if ((empty || overwrite) && values[i] != "") {
                ComponentFactory.setComponentValue(
                        o.getFieldType(destinationFields[i]),
                        cInputFields[destinationFields[i]],
                        values[i]);
            }
        }
    }    

    /**
     * Returns the data from screen
     *
     * @return (Object[])  the data as array
     */
    private Object[] getData() {
        Object[] values = o.getValues();
        for (int i = 0; i < cInputFields.length; i++) {
            Object value = ComponentFactory.getComponentValue(
                            o.getFieldType(i), 
                            cInputFields[i]);
            values[i] = value;
        }
        
        if (!update) {
            values[DCObject._ID] = "";
        }
        
        return values;
    }

    /**
     * Starts a Amazon update. The update will show its own dialog
     * and transfers data selected by the user back to this record
     * form.
     */
    public void startAmazonUpdate() {
        Object[] data = this.getData();
        Object oName = data[Software._A_NAME];
        Object oID = null;
        
        if (oName != null) {
            Object[][] o = {{oID, oName}};
            Messages.setMessage(Resources.getText(DCMessages.msgAmazonUpdateStarts), 
                                Messages._STANDARD);
            Thread threadAmazon = new Thread(new AmazonUpdate(o, this));
            threadAmazon.setDaemon(true);
            threadAmazon.setPriority(1);
            threadAmazon.start();            
        } else {
            new MessageBox(Resources.getText(DCMessages.msgAmazonNeedsTitle),
                           MessageBox._WARNING);
            return;   
        } 
    }

    /**
     * Starts a IMDB update. The update will show its own dialog
     * and transfers data selected by the user back to this record
     * form.
     */
    public void startIMDBUpdate() {
        Object[] data = this.getData();
        Object oTitle = data[Movie._A_TITLE];
        Object oID = null;
        
        if (oTitle != null) {
            Object[][] o = {{oID, oTitle}};
            Thread threadIMDB = new Thread(new ImdbUpdate(o, this));
            threadIMDB.setDaemon(true);
            threadIMDB.setPriority(1);
            threadIMDB.start();            
        } else {
            new MessageBox(Resources.getText(DCMessages.msgIMDBNeedsTitle),
                           MessageBox._WARNING);
            return;   
        } 
    }

    /**
     * Deletes this record from the database
     */
    protected void deleteValues() {
        new QuestionBox(Resources.getText(DCMessages.msgDeleteQuestion), this);
        
        if (QuestionBox.result.booleanValue()) {
            String id = o.getValue(DCObject._ID).toString();
            o.clearValues();
            o.setValue(DCObject._ID, id);
            o.setSilence(true);
            
            Object[] params = {this};
            Requestor requestClose = new Requestor(TablePanel.class, "closeRecord", params);
            RequestorCollection requestors = new RequestorCollection();
            requestors.addRequestor(requestClose);
    
            Requestor requestor = new Requestor(clazz, 
                                                "removeRows", 
                                                new Object[] {(new int[] {row})});
            requestors.addRequestor(requestor);
            o.delete(requestors);
        }
    }

    /**
     * Inserts or updates the Data Crow object
     */
    protected void saveValues() {
        String id = o.getValue(DCObject._ID).toString();
        
        Object[] params = {this};
        Requestor requestClose = new Requestor(TablePanel.class, "closeRecord", params);
        RequestorCollection requestors = new RequestorCollection();
        requestors.addRequestor(requestClose);

        Object[] newVals = getData();
        for (int i = 0; i < newVals.length; i++) {
            if (newVals[i] == null) {
                newVals[i] = "";
            }
        }        
        
        o.setSilence(true);
        if (!update)  {
            // Just insert the data into the database
            o.setValues(newVals);
            o.setValue(DCObject._ID, "");
            o.saveNew(requestors);
        } else {
            // get the orginal values
            Object[] oldVals = o.getValues();
            
            // check if values have been changed and if so update
            // them from the UI into the Data Crow Object. This will
            // mark the field as being changed on the Data Crow Object
            for (int i = 0; i < newVals.length; i++) {
                if (oldVals[i] != null && 
                    !oldVals[i].toString().equals(newVals[i].toString())) {
                    o.setValue(i, newVals[i]);
                }
            }
            
            if (o.isChanged()) {
                Requestor requestor = new Requestor(clazz, 
                                                    "updateRow", 
                                                    new Object[] {id, newVals});
                requestors.addRequestor(requestor);
                o.saveUpdate(requestors);
            }
        }
    }

    /**
     * Closes this form by setting the visibility to false
     */
    public void close() {
        SwingUtil.setRootFrame(null);

        clazz = null;
        
        Settings.setValue(getRecordSizeSetting(), getSize());
        o = null;
        dispose();
    }

    private class IMDBAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            startIMDBUpdate();
        }
    }

    private class AmazonAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            startAmazonUpdate();
        }
    }

    private class SaveAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            saveValues();
        }
    }

    private class DeleteAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            deleteValues();
        }
    }

    private class ExitAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            close();
        }
    }

    /**
     * Forces a refresh of the content of the Software categories combobox
     */
    public class ActiveFrame implements MouseListener {
        public void mouseReleased(MouseEvent e) {}
        public void mouseEntered(MouseEvent e) {
            SwingUtil.setRootFrame((JFrame) e.getSource());
        }
        public void mouseExited(MouseEvent e) {}
        public void mousePressed(MouseEvent e) {}
        public void mouseClicked(MouseEvent e) {}        
    }  
}
