/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 Robert-Jan van der Waals                    *
 *                             rwaals@anytimenow.com                          *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.console;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.Image;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.LinkedList;

import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;

import net.sf.dc.console.components.ComponentFactory;
import net.sf.dc.console.components.DCTableModel;
import net.sf.dc.console.components.Layout;
import net.sf.dc.console.movie.ImdbMovieSearchForm;
import net.sf.dc.console.movie.InsertMoviePanel;
import net.sf.dc.console.movie.SearchMoviePanel;
import net.sf.dc.console.software.AmazonSoftwareSearchForm;
import net.sf.dc.console.software.InsertSoftwarePanel;
import net.sf.dc.console.software.SearchSoftwarePanel;
import net.sf.dc.core.IconLibrary;
import net.sf.dc.core.Repository;
import net.sf.dc.core.modules.DCModules;
import net.sf.dc.core.objects.DCObject;
import net.sf.dc.core.objects.Requestor;
import net.sf.dc.core.objects.RequestorCollection;
import net.sf.dc.core.objects.movie.Movie;
import net.sf.dc.core.objects.software.Software;
import net.sf.dc.resources.DCLabels;
import net.sf.dc.resources.DCMessages;
import net.sf.dc.resources.DCTooltips;
import net.sf.dc.wf.WorkFlow;
import net.sf.ext.jflexsett.settings.Settings;
import net.sf.ext.util.Resources;
import net.sf.ext.util.SwingUtil;
import net.sf.ext.util.Utilities;


/**
 * Base for all search forms. Used for internet updates
 *
 * @author Robert-Jan van der Waals
 * @since 1.3
 * @version DC0.50
 */
public abstract class SearchForm extends JFrame {
    private boolean overwrite = Settings.getValueAsBoolean(
                                Repository.settings.stODBOverwriteValues);
    private boolean broadband = Settings.getValueAsBoolean(
                                Repository.settings.stBroadband);
    
    protected JTextField textSearch;
    protected JCheckBox checkOverwrite;
    protected JCheckBox checkBroadband;

    protected Object sender;    
    protected Thread threadSearch;

    protected JProgressBar progress;
    protected JButton buttonSearch;
    protected JButton buttonStop;  
    protected JButton buttonStopAll;

    // indicates if a process is running 
    protected boolean running = false;
    protected boolean continueNext = true;

    protected final DCTableModel tableModel;
    protected JTable tableResult;
    protected int rowNumber = -1;
    
    protected int[] fieldIndices;

    protected boolean isUpdateRunning = false;
    
    /**
     * Opens the search form and build the dialog
     * 
     * @param sender the requestor
     */
    public SearchForm(Object sender, int[] fieldIndices, int module) {
        SwingUtil.setRootFrame(this);
        
        tableModel = new DCTableModel(DCModules.getModule(module));
        
        Image image = IconLibrary._icoInternet.getImage();
        this.setIconImage(image);
        image.flush();
        this.sender = sender;
        this.fieldIndices = fieldIndices;
        buildDialog(); 
        
        Dimension location = Utilities.getCenteredWindowLocation(getSize());
        setLocation(location.width, location.height);
    }
    
    /**
     * Terminates all actions
     */
    public void stop() {
        Settings.setValue(Repository.settings.stODBOverwriteValues,
                          new Boolean(checkOverwrite.isSelected()));
        Settings.setValue(Repository.settings.stBroadband,
                          new Boolean(checkBroadband.isSelected()));
        Settings.setValue(Repository.settings.stImdbSearchFormViewSize,
                          getSize());
        isUpdateRunning = false;

        running = false;
        progress.setValue(0);
        dispose();
    }

    /**
     * Must be implemented by other classes; starts the search
     * actions and fills this screen with the results
     */
    protected abstract void startSearch();

    /**
     * Indicates if an update is running or not 
     */
    public boolean isUpdateRunning() {
        return isUpdateRunning;   
    } 
    
    /**
     * Indicates if the user asked to stop the whole update process
     *
     */
    public boolean continueNext() {
        return continueNext;   
    }

    /**
     * Adds a result as a row to the table
     * 
     * @param data one row
     */
    public void setData(DCObject obj) {
        tableModel.addRow(obj.getValues());
    }
    
    /**
     * Transfers the selected result to the requestor 
     */
    protected void transferToRequestor() {
        Object[] row = getSelectedData();

        if (row == null) {
            return;
        }
        
        Class clazz = null;
        String sMethod = "updateRowAt";
        if (sender instanceof InsertMoviePanel) {
            clazz = InsertMoviePanel.class;    
        } else if (sender instanceof SearchMoviePanel) {
            clazz = SearchMoviePanel.class;   
        } else if (sender instanceof SearchSoftwarePanel) {
            clazz = SearchSoftwarePanel.class;   
        } else if (sender instanceof InsertSoftwarePanel) {
            clazz = InsertSoftwarePanel.class;  
        } else {
            clazz = RecordForm.class; 
            sMethod = "setData";
        }
        
        Object[] params = null;
        RequestorCollection requestors = new RequestorCollection();

        if (clazz.getName().equals(RecordForm.class.getName())) {
            Object[] o = {sender, 
                          row, 
                          new Boolean(checkOverwrite.isSelected()),
                          fieldIndices};
            params = o;
        } else {
            Object[] o = {new Integer(rowNumber), 
                          row, 
                          new Boolean(checkOverwrite.isSelected()),
                          fieldIndices};
            params = o;
        }
                               
        Requestor request = new Requestor(clazz, 
                                          sMethod, 
                                          params);
                                          
        requestors.addRequestor(request);
        WorkFlow wf = new WorkFlow();
        wf.displayResults(new LinkedList(), requestors, true);
        stop();
    }
    
    /**
     * Returns the selected data (in the table)
     */
    public Object[] getSelectedData() {
        int row = tableResult.getSelectedRow();
        
        if (row < 0) {
            new MessageBox(Resources.getText(DCMessages.msgSelectRowForTransfer),
                           MessageBox._WARNING);
            return null;
        }
        
        Object[] o = new Object[tableResult.getColumnCount()];
        for (int i = 0; i < o.length; i++) {
            o[i] = tableResult.getValueAt(row, i);
        }
        return o;
    }
    
    public void buildDialog() {
        Color color = Settings.getValueAsColor(Repository.settings.stColor);
        this.getContentPane().setBackground(color);
        JPanel actionPanel = new JPanel(false);
        actionPanel.setBackground(color);
        JPanel resultPanel = new JPanel(false);
        resultPanel.setBackground(color);        
        JPanel searchPanel = new JPanel(false);
        searchPanel.setBackground(color);
        JPanel progressPanel = new JPanel(false);
        progressPanel.setBackground(color);
        
        // The search part
        textSearch = ComponentFactory.getTextField();
        buttonSearch =  ComponentFactory.getButton(
                        IconLibrary._icoFetch,
                        Resources.getText(DCLabels.lblSearch));
        buttonSearch.addActionListener(new SearchAction());
        
        searchPanel.setLayout(Layout.getGBL());                                
        searchPanel.add(textSearch,   Layout.getGBC( 0, 0, 1, 1, 90.0, 90.0
                                     ,GridBagConstraints.WEST, GridBagConstraints.HORIZONTAL,
                                      new Insets(5, 5, 5, 5), 0, 0));                         
        searchPanel.add(buttonSearch, Layout.getGBC( 1, 0, 1, 1, 1.0, 1.0
                                     ,GridBagConstraints.EAST, GridBagConstraints.HORIZONTAL,
                                      new Insets(5, 5, 5, 5), 0, 0));

        // Action panel
        JButton buttonTransfer = ComponentFactory.getButton(
                                 Resources.getText(DCLabels.lblTransfer));
        buttonStop = ComponentFactory.getButton(
                     Resources.getText(DCLabels.lblStop));
        buttonStopAll = ComponentFactory.getButton(
                        Resources.getText(DCLabels.lblStopAll));
        JButton buttonClose = ComponentFactory.getButton(
                              Resources.getText(DCLabels.lblClose));
        
        buttonStopAll.setToolTipText(Resources.getText(DCTooltips.tpStopAll));
        buttonStop.setToolTipText(Resources.getText(DCTooltips.tpStopSearch));
        buttonClose.setToolTipText(Resources.getText(DCTooltips.tpClose));
        buttonTransfer.setToolTipText(Resources.getText(DCTooltips.tpTransfer));
        
        buttonTransfer.addActionListener(new TransferRow());
        buttonClose.addActionListener(new CloseWindow());
        buttonStop.addActionListener(new StopAction());
        buttonStopAll.addActionListener(new StopAllAction());

        actionPanel.add(buttonTransfer);
        actionPanel.add(buttonStop);
        actionPanel.add(buttonStopAll);
        actionPanel.add(buttonClose);
        
        // Result panel
        // get the default movie table and remove the unneeded rows (location and such)
        if (this.getClass().getName().equals(ImdbMovieSearchForm.class.getName())) {
            tableResult = ComponentFactory.getTableForObject(new Movie(), tableModel);

            tableResult.getColumnModel().removeColumn(tableResult.getColumnModel().getColumn(24));
            tableResult.getColumnModel().removeColumn(tableResult.getColumnModel().getColumn(23));
            tableResult.getColumnModel().removeColumn(tableResult.getColumnModel().getColumn(22));
            tableResult.getColumnModel().removeColumn(tableResult.getColumnModel().getColumn(20));
            tableResult.getColumnModel().removeColumn(tableResult.getColumnModel().getColumn(19));
            tableResult.getColumnModel().removeColumn(tableResult.getColumnModel().getColumn(18));
            tableResult.getColumnModel().removeColumn(tableResult.getColumnModel().getColumn(17));
            tableResult.getColumnModel().removeColumn(tableResult.getColumnModel().getColumn(16));
            tableResult.getColumnModel().removeColumn(tableResult.getColumnModel().getColumn(15));
            tableResult.getColumnModel().removeColumn(tableResult.getColumnModel().getColumn(14));
            tableResult.getColumnModel().removeColumn(tableResult.getColumnModel().getColumn(13));
            tableResult.getColumnModel().removeColumn(tableResult.getColumnModel().getColumn(12));
            tableResult.getColumnModel().removeColumn(tableResult.getColumnModel().getColumn(11));
            tableResult.getColumnModel().removeColumn(tableResult.getColumnModel().getColumn(10));
            tableResult.getColumnModel().removeColumn(tableResult.getColumnModel().getColumn(0));
        } else if (this.getClass().getName().equals(AmazonSoftwareSearchForm.class.getName())) {
            tableResult = ComponentFactory.getTableForObject(new Software(), tableModel);
            tableResult.getColumnModel().removeColumn(tableResult.getColumnModel().getColumn(13));
            tableResult.getColumnModel().removeColumn(tableResult.getColumnModel().getColumn(9));
            tableResult.getColumnModel().removeColumn(tableResult.getColumnModel().getColumn(8));
            tableResult.getColumnModel().removeColumn(tableResult.getColumnModel().getColumn(6));
            tableResult.getColumnModel().removeColumn(tableResult.getColumnModel().getColumn(5));
            tableResult.getColumnModel().removeColumn(tableResult.getColumnModel().getColumn(4));
            tableResult.getColumnModel().removeColumn(tableResult.getColumnModel().getColumn(0)); 
        }

        for (int i = 0; i < tableResult.getColumnCount(); i++) {
        	tableResult.getColumnModel().getColumn(i).setHeaderRenderer(
                    ComponentFactory.getTableHeaderRenderer());   
        }
        
        JScrollPane scrollResults = new JScrollPane(tableResult);
        scrollResults.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);        

        checkOverwrite = ComponentFactory.getCheckBox(
                         Resources.getText(DCLabels.lblOverwriteExistingValues));
        checkOverwrite.setBackground(color);
        checkOverwrite.setToolTipText(Resources.getText(DCTooltips.tpOverwriteExistingValues));
        
        checkBroadband = ComponentFactory.getCheckBox(
        				 Resources.getText(DCLabels.lblBroadbandConnection));
        checkBroadband.setBackground(color);

        resultPanel.setLayout(Layout.getGBL());
        resultPanel.add(scrollResults,    Layout.getGBC( 0, 0, 1, 1, 90.0, 90.0
                                         ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
                                          new Insets(5, 5, 5, 5), 0, 0));   
        resultPanel.add(checkOverwrite,   Layout.getGBC( 0, 1, 1, 1, 1.0, 1.0
                                         ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                                          new Insets(5, 5, 5, 5), 0, 0));   
        resultPanel.add(checkBroadband,   Layout.getGBC( 0, 2, 1, 1, 1.0, 1.0
                						 ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
										  new Insets(5, 5, 5, 5), 0, 0));   
        
        // Progress panel
        progress = new JProgressBar();
        progressPanel.setLayout(Layout.getGBL());
        progressPanel.add(progress, Layout.getGBC( 0, 0, 1, 1, 1.0, 1.0
                         ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                          new Insets(5, 5, 5, 5), 0, 0));

        // Main panel                                          
        this.getContentPane().setLayout(Layout.getGBL());
        this.getContentPane().add(searchPanel, Layout.getGBC( 0, 0, 1, 1, 1.0, 1.0
                             ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                              new Insets(5, 5, 5, 5), 0, 0)); 
        this.getContentPane().add(resultPanel, Layout.getGBC( 0, 1, 1, 1, 90.0, 90.0
                             ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
                              new Insets(5, 5, 5, 5), 0, 0)); 
        this.getContentPane().add(actionPanel, Layout.getGBC( 0, 2, 1, 1, 0.0, 0.0
                             ,GridBagConstraints.SOUTHWEST, GridBagConstraints.NONE,
                              new Insets(5, 5, 5, 5), 0, 0)); 
        this.getContentPane().add(progressPanel,Layout.getGBC(0, 3, 1, 1, 1.0, 1.0
                             ,GridBagConstraints.SOUTHWEST, GridBagConstraints.HORIZONTAL,
                              new Insets(5, 5, 5, 5), 0, 0)); 
        
        checkOverwrite.setSelected(overwrite);
        checkBroadband.setSelected(broadband);
        
        this.setResizable(true);
        this.pack(); 
        buttonStop.setEnabled(false);
        buttonStopAll.setEnabled(false);
        this.setSize(Settings.getValueAsDimension(Repository.settings.stImdbSearchFormViewSize));
    }
    
    private class SearchAction implements ActionListener{
        public void actionPerformed(ActionEvent e) {
            startSearch();   
        }
    }
    
    private class TransferRow implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            transferToRequestor();   
        }
    }

    private class CloseWindow implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            stop();
        }
    }
    
    private class StopAllAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            continueNext = false;
            running = false;
            buttonSearch.setEnabled(true);
            buttonStop.setEnabled(false);
            buttonStopAll.setEnabled(false);
        }
    }     
    
    private class StopAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            running = false;
            buttonSearch.setEnabled(true);
            buttonStop.setEnabled(false);
            buttonStopAll.setEnabled(false);
        }
    }    
}
