/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 Robert-Jan van der Waals                    *
 *                             rwaals@anytimenow.com                          *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.console;

import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.Vector;

import javax.swing.JButton;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JSeparator;
import javax.swing.JTable;
import javax.swing.SwingUtilities;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.TableColumn;

import net.sf.dc.console.components.ComponentFactory;
import net.sf.dc.console.components.DCTableCellRenderer;
import net.sf.dc.console.components.DCTableModel;
import net.sf.dc.console.components.Layout;
import net.sf.dc.core.DataCrow;
import net.sf.dc.core.modules.DCModule;
import net.sf.dc.core.modules.DCModules;
import net.sf.dc.core.objects.DCObject;
import net.sf.dc.export.TableToFileExporter;
import net.sf.dc.messages.Messages;
import net.sf.dc.processes.AmazonUpdate;
import net.sf.dc.processes.ImdbUpdate;
import net.sf.dc.resources.DCLabels;
import net.sf.dc.resources.DCMessages;
import net.sf.dc.resources.DCTooltips;
import net.sf.ext.jflexsett.settings.Settings;
import net.sf.ext.util.Resources;
import net.sf.ext.util.StringUtils;

/**
 * Base panel for all panels which contain a JTable
 * 
 * This class facilities most of the functions needed to fill / clear /
 * update tables. It also provides a popup menu for the table and 
 * action buttons.
 *
 * @author Robert-Jan van der Waals
 * @since 1.3
 * @version DC0.50
 */ 
public abstract class TablePanel extends DataCrowPanel {

	protected final DCTableModel tableModel;
    protected JTable table;
    protected TableActionsPanel panelTableActions;
    protected PopupMenu popupMenu;

    // holds only the changed objects. key = ID, value = DCObject
    protected final Hashtable cache = new Hashtable();
    
    protected final TableValueChangedAction tableChangeListener = new TableValueChangedAction();
    protected final TableMouseListener tableMouseListener = new TableMouseListener();
    
    protected Vector vHiddenColumns = new Vector();
    protected Vector vColumns = new Vector();

    protected int moduleIndex;

    public TablePanel(int module) {
        super();  
        this.moduleIndex = module;
        
        if (moduleIndex != -1) {
        	tableModel = new DCTableModel(DCModules.getModule(module));
        } else {
        	tableModel = new DCTableModel(null);
        }
    }
    
    /**
     * Tries to apply the current editing of the table cell.
     * If the current value cannot be applied, the editing is simply
     * cancelled (changes are discarded) 
     */
    public void cancelEdit() {
        int selectedRow = table.getSelectedRow();
        int selectedColumn = table.getSelectedColumn();
        
        if (selectedRow != -1) {
            
            
            for (int i = 0; i < table.getColumnCount(); i++) {
                boolean success = false;
                try {
                    table.getCellEditor(selectedRow, i).stopCellEditing();
                    success = true;
                } catch (Exception ignore) {}
                
                if (!success) {
                    table.getCellEditor(selectedRow, i).cancelCellEditing();
                }
            }
        }
    }

    /**
     * Retrieves the module to which this panel belongs 
     */
    protected DCModule getModule() {
    	return DCModules.getModule(moduleIndex);
    }
    
    /**
     * Stores the current hidden fields indices in the Settings.
     * @param modelIndices the table model indices / field indices which are
     *        marked as hidden 
     */
    private void setHiddenFieldsSetting(int[] modelIndices) {
        Settings.setValue(getModule().getHiddenFieldsSettingsKey(), 
                          modelIndices);
    }
    
    /**
     * Opens the fields selector dialog and applies the hidden fields on the
     * table. Also the hidden fields setting is set for the current module.
     */
    public void startFieldSelector() {
        Vector vVisibleColumns = new Vector();
        for (Enumeration enum = table.getColumnModel().getColumns(); enum.hasMoreElements(); ) {
            vVisibleColumns.addElement(enum.nextElement());
        }
        FieldSelectorDialog dialog = new FieldSelectorDialog(vVisibleColumns, 
                                                             vHiddenColumns);
        dialog.setVisible(true);
        if (FieldSelectorDialog.isSuccessful()) {
            vVisibleColumns = dialog.getVisibleFields(); 
            vHiddenColumns = dialog.getHiddenFields();

            // get the hidden columns table model indices by comparing the
            // columns of the table model with the table              
            int[] modelIndices = new int[vHiddenColumns.size()];
            int counter = 0;
            for (Enumeration enum = vHiddenColumns.elements(); enum.hasMoreElements(); ) {
                TableColumn column = (TableColumn) enum.nextElement();
                modelIndices[counter++] = column.getModelIndex();
            }            
            
            setHiddenFieldsSetting(modelIndices);
            MainFrame.setHiddenFields();
        }
    }
    
    /**
     * Adds all the columns to the table (hidden and visible fields)
     */
    private void resetTable() {
        try {
            for (Enumeration enum = vColumns.elements(); enum.hasMoreElements(); ) {
                TableColumn column = (TableColumn) enum.nextElement();
                
                try {
                    table.getColumn(column.getIdentifier());
                } catch (Exception exp) {
                    table.addColumn(column);
                }
            }
        } catch (Exception ignore) {}
    }
    
    /**
     * Removes the hidden fields from the view.
     * 
     * @param hiddenFields the hidden fields to be removed
     */
    private void removeHiddenFields(Vector hiddenFields) {
        try {
            for (Enumeration enum = hiddenFields.elements(); enum.hasMoreElements(); ) {
                TableColumn column = (TableColumn) enum.nextElement();
                table.removeColumn(column);
            }
        } catch (Exception ignore) {}
        setRequiredFields(getModule().getDataCrowObject());
    }
    
    /**
     * External use. All columns to be removed are still visible
     * in the table. The columns are retrieved, stored in a Vector 
     * and then removed from the view
     * 
     * @param hiddenColumns The model indices of the columns
     */
    public void removeHiddenFields(int[] hiddenColumns) {
        cancelEdit();
        resetTable();
        
        vHiddenColumns.clear();
        
        for (int i = 0; i < hiddenColumns.length; i++) {
            int tableIndex = table.convertColumnIndexToView(hiddenColumns[i]);
            TableColumn column = table.getColumnModel().getColumn(tableIndex);
            vHiddenColumns.addElement(column); 
        }
        removeHiddenFields(vHiddenColumns);   
    }
    
    /**
     * Fills the table with the specified values
     *
     * @param data list containing the rows of data to be displayed
     */
    public void fillTable(LinkedList data) {
        int selectedRow = table.getSelectedRow();
        int selectedCol = table.getSelectedColumn();
        
        clearTable();

        // clear the tablemodel of data as well
        for (   Enumeration enum = tableModel.getDataVector().elements(); 
                enum.hasMoreElements(); ) {
            ((Vector) enum.nextElement()).clear();
        }
        tableModel.getDataVector().clear();
        
        try {
            while (data.size() > 0) {
                DCObject obj = (DCObject) data.removeFirst();
                Object[] values = obj.getValues();
                tableModel.addRow(values);
                values = null;
                obj = null;
            }
            String sRecords = "" + tableModel.getRowCount();
            setStatus(Resources.getText(DCMessages.msgSearchResult, sRecords));
        } catch (java.lang.Exception exp) {
            Messages.setMessage(exp);
        }
        
        try {
            table.setRowSelectionInterval(selectedRow, selectedRow);
            table.setColumnSelectionInterval(selectedCol, selectedCol);            
        } catch (Exception ignore) {}
        data.clear();
        data = null;
    }

    /**
     * Returns the row number of the record with the given ID
     * (product object ID; ie. the value of Software._ID)
     *
     * @param ID
     */
    public int getRowNumberWithID(String ID) {
        cancelEdit();
        for (int i = 0; i < tableModel.getRowCount(); i++) {
            String sID = tableModel.getValueAt(i, 0).toString();
            if (sID.equals(ID)) {
                return i;
            }
        }
        return -1;
    }

    /**
     * Updates the row with the given ID with the data (values)
     *
     * @param ID the ID (Product Object ID)
     * @param values the new values
     */
    public void updateRow(String ID, Object[] values) {
        cancelEdit();

        int row = getRowNumberWithID(ID);
        if (row != -1) {
            for (int j = 0; j < values.length; j++) {
                tableModel.setValueAt(values[j], row, j);
                addRowToCache(row, j);
            }
        }
    }

    /**
     * Returns all the values at the give row
     * 
     * @param rowNumber the UI row number
     * @param hidden indicates if values of hidden fields should also be returned 
     */
    public Object[] getValuesAtRow(int rowNumber, boolean hidden) {
        cancelEdit();
        
        Object[] row = null;
        if (hidden) {
            row = new Object[tableModel.getColumnCount()];
        } else {
            row = new Object[table.getColumnCount()];
        }
        for (int j = 0; j < row.length; j++) {
            Object oTest = null;
            if (hidden) {
                oTest = tableModel.getValueAt(rowNumber, j);
            } else {
                oTest = table.getValueAt(rowNumber, j);    
            }
            if (oTest != null) {
                row[j] = oTest.toString();
            } else {
                row[j] = "";
            }                
        }
        return row;
    }

    /**
     * Adds the datacrow objects to the table
     *
     * @param vDataCrowObjects  vector of datacrow objects
     */
    public void addRow(LinkedList data) {
        while (data.size() > 0) {
            DCObject lcObj = (DCObject) data.removeFirst();
            Object[] row = lcObj.getValues();
            row[0] = "" + (tableModel.getRowCount() + 1);
            tableModel.addRow(row);
            lcObj = null;
            row = null;
        }
        data.clear();
        data = null;
    }

    /**
     * Clear the table of all values
     */
    public void clearTable() {
        cache.clear();
        cancelEdit();
        
        tableModel.removeTableModelListener(tableChangeListener);
        tableModel.setRowCount(0);
        tableModel.addTableModelListener(tableChangeListener);
        setStatus(Resources.getText(DCMessages.msgTableCleared));
    }

    /**
     * Updates a row at the given rownumber. If indicated this will
     * update existing data.
     * 
     * @param rowNumber UI row to be updated
     * @param rowData the data to be written to the row
     * @param overWrite indicates of existing data should be 
     *        overwritten
     * @param fieldIndices the field indices (communicates
     *        directly to the column indices) 
     */
    public void updateRowAt(int rowNumber,
                            Object[] rowData,
                            boolean overWrite,
                            int[] fieldIndices) {
        cancelEdit();

        for (int i = 0; i < rowData.length; i++) {
            Object oNew = rowData[i];
            Object oExist = tableModel.getValueAt(rowNumber, fieldIndices[i]);

            if (!overWrite && (oExist != null &&
                               !oExist.toString().trim().equals(""))) {
                oNew = oExist;
            }

            if (oNew == null || oNew.toString().trim().equals("")) {
                oNew = oExist;
            }
            tableModel.setValueAt(oNew, rowNumber, fieldIndices[i]);
            addRowToCache(rowNumber, fieldIndices[i]);
        }
    }

    /**
     * Gathers the data for an internet update
     * 
     * @param rows the row indices of the to be updated rows
     */
    private Object[][] getDataForInternetUpdate(int[] rows) {
        cancelEdit();
        Vector vNames = new Vector(0);
        Vector vIDs = new Vector(0);

        if (rows == null || rows.length < 1) {
            new MessageBox(Resources.getText(DCMessages.msgSelectRowUpdate),
                           MessageBox._WARNING);
            return null;
        }

        for (int i = 0; i < rows.length; i++) {
            Object oName = tableModel.getValueAt(rows[i], 1);
            oName = oName == null ? "" : oName;
            vNames.add(oName);
            vIDs.add(new Integer(rows[i]));
        }

        Object[][] o = new Object[vNames.size()][2];
        for (int j = 0; j < o.length; j++) {
            o[j][0] = vIDs.elementAt(j);
            o[j][1] = vNames.elementAt(j);
        }

        return o;
    }
    
    /**
     * Starts the Amazon internet update. Th ejob is run 
     * in a seperate thread
     */
    public void startAmazonUpdate() {
        cancelEdit();
        Messages.setMessage(Resources.getText(DCMessages.msgAmazonUpdateStarts),
                            Messages._STANDARD);

        Object[][] o = getDataForInternetUpdate(table.getSelectedRows());

        Thread threadAmazon = new Thread(new AmazonUpdate(o, this));
        threadAmazon.setDaemon(true);
        threadAmazon.setPriority(1);
        threadAmazon.start();
    }

    
    /**
     * Starts the IMDB internet update. The job is run 
     * in a seperate thread
     */
    public void startIMDBUpdate() {
        cancelEdit();

        Object[][] o = getDataForInternetUpdate(table.getSelectedRows());

        Thread threadIMDB = new Thread(new ImdbUpdate(o, this));
        threadIMDB.setDaemon(true);
        threadIMDB.setPriority(1);
        threadIMDB.start();
    }

    /**
     * Opens a record form for the selected record
     * 
     * 1) Cancel all edits. When in the middle of updating a cell, the new value
     *    will be applied
     * 2) Create a representation of the row. Values of hidden fields are also 
     *    retrieved
     * 3) Retrieve the olf values from the cache
     * 4) Check for each value if it has been changed and if so, update it
     */
    public void openRecord() {
        // 1
        cancelEdit();
        int selectedRow = table.getSelectedRow();
        if (selectedRow == -1) {
            new MessageBox(Resources.getText(DCMessages.msgSelectRowToOpen),
                           MessageBox._WARNING);
            return;
        }

        // 2
        Object[] row = new Object[tableModel.getColumnCount()];
        for (int i = 0; i < row.length; i++) {
            Object oSelected = tableModel.getValueAt(selectedRow, i);
            if (oSelected == null) {
                row[i] = "";
            } else {
                row[i] = oSelected;
            }
        }

        DCObject obj;
        String sID = tableModel.getValueAt(selectedRow, 0).toString();
        
        if (cache.containsKey(sID)) {
        	obj = (DCObject) cache.get(sID);
        } else {
            obj = getModule().getDataCrowObject();
            obj.setValues(getValuesAtRow(table.getSelectedRow(), true));
        }

        boolean update = this instanceof TableSearchPanel ? true : false;

        new RecordForm( update,
                        obj,
                        getModule().isImdbUpdateAllowed(),
                        getModule().isAmazonUpdateAllowed(),
                        getClass(),
                        selectedRow);
    }

    /**
     * Remove the rows from the table with the given indices
     */
    public void removeRows(int[] rows) {
        cancelEdit();
        for (int i = rows.length -1; i > -1; i--) {
            tableModel.removeRow(rows[i]);
        }
    }

    /**
     * Creates a visual representation (different header) 
     * of the required fields
     * 
     * @param o data crow object (used as a reference)
     */
    public void setRequiredFields(DCObject o) {
        int counter = 0; 
        for (Enumeration enum = vColumns.elements(); enum.hasMoreElements(); ) {
            TableColumn column = (TableColumn) enum.nextElement();
            if (o.isRequired(counter)) {
                column.setHeaderRenderer(
                    ComponentFactory.getRequiredTableHeaderRenderer());
            } else {
                column.setHeaderRenderer(
                    ComponentFactory.getTableHeaderRenderer());
            }
            counter++;
        }
    }

    /**
     * Returns all Data Crow Objects currently persisting in the
     * (UI) table
     */
    private DCObject[] getData() {
        DCObject[] objects = new DCObject[tableModel.getRowCount()];
    	for (int i = 0; i < objects.length; i++) {
    		Object[] data = getValuesAtRow(i, true);
            objects[i] = getModule().getDataCrowObject();
            objects[i].setValues(data);
        }
        return objects;
    }
    
    /**
     * Creates a PDF report. All visible fields are reported on.
     * A file browser is opened in which the user can indicate where
     * the pdf file should be stored.
     */
    public void createPdfReport() {
    	new TableToFileExporter(getData(), TableToFileExporter._PDF);
    }

    /**
     * Menu shown when the user performs a right click on a table.
     * The menu is build up from information from the owner module.
     *  
     * @author rjwaal
     * @since 1.3
     * @version DC1.00
     */
    public class PopupMenu extends JPopupMenu {
        JMenuItem menuOpen;
        JMenuItem menuSave;
        JMenuItem menuDelete;
        JMenuItem menuUpdateAll;
        JMenuItem menuIMDB;
        JMenuItem menuAmazon;
        JMenuItem menuRemoveRow;
        JMenuItem menuAddRow;
        
        JMenuItem menuReportPdf;
        
        JMenuItem menuFieldSelector;

        public PopupMenu(boolean showImdb,
                         boolean showAmazon,
                         boolean showDelete,
                         boolean showAddRemove,
                         boolean showOpen,
                         boolean showExport) {

            menuOpen = new JMenuItem(Resources.getText(DCLabels.lblOpenRecord));
            menuSave = new JMenuItem(Resources.getText(DCLabels.lblSave));
            menuDelete = new JMenuItem(Resources.getText(DCLabels.lblDelete));
            menuIMDB = new JMenuItem(Resources.getText(DCLabels.lblIMDBUpdate));
            menuAmazon = new JMenuItem(Resources.getText(DCLabels.lblAmazonUpdate));
            menuRemoveRow = new JMenuItem(Resources.getText(DCLabels.lblRemoveRow));
            menuAddRow = new JMenuItem(Resources.getText(DCLabels.lblAddRow));
            menuReportPdf = new JMenuItem(Resources.getText(DCLabels.lblExportToPDF));
            menuFieldSelector = new JMenuItem(Resources.getText(DCLabels.lblShowFieldSelector));
            menuUpdateAll = new JMenuItem(Resources.getText(DCLabels.lblUpdateAll));

            menuFieldSelector.addActionListener(new FieldSelectorAction());
            menuOpen.addActionListener(new OpenRecordAction());
            menuRemoveRow.addActionListener(new RemoveRowAction());
            menuAddRow.addActionListener(new AddRowAction());
            
            menuReportPdf.addActionListener(new CreatePdfReport());
            
            menuUpdateAll.addActionListener(new UpdateAllAction());

            this.add(menuFieldSelector);
            this.add(menuUpdateAll);
            this.add(new JSeparator(JSeparator.HORIZONTAL));

            if (showOpen) {
                this.add(menuOpen);
            }
            
            this.add(menuSave);
            if (showDelete) {
                this.add(menuDelete);
            }
            if (showAddRemove) {
                this.add(menuRemoveRow);
                this.add(menuAddRow);
            }
            
            if (showImdb) {
                this.add(new JSeparator(JSeparator.HORIZONTAL));
                this.add(menuIMDB);
            }

            if (showAmazon) {
                this.add(new JSeparator(JSeparator.HORIZONTAL));
                this.add(menuAmazon);
            }
            
            if (showExport) {
                this.add(new JSeparator(JSeparator.HORIZONTAL));
                this.add(menuReportPdf);
            }
        }

        public void addAmazonActionListener(ActionListener listener) {
            menuAmazon.addActionListener(listener);
        }

        public void addDeleteActionListener(ActionListener listener) {
            menuDelete.addActionListener(listener);
        }

        public void addIMDBActionListener(ActionListener listener) {
            menuIMDB.addActionListener(listener);
        }

        public void addSaveActionListener(ActionListener listener) {
            menuSave.addActionListener(listener);
        }
    }

    private void addRowToCache(int row, int column) {
    	tableModel.removeTableModelListener(tableChangeListener);
        
        if (row != -1 && column != -1) {
            String sID = tableModel.getValueAt(row, 0).toString();
            DCObject o;
            if (!cache.containsKey(sID)) {
                o = getModule().getDataCrowObject();
                
                Object[] values = getValuesAtRow(row, true);
                o.setValues(values);
                values = null;
                o.undoChanges();
            } else {
                o = (DCObject) cache.get(sID);
            }
            o.setValue(column, tableModel.getValueAt(row, column));
            cache.put(sID, o);
        }        
        
        tableModel.addTableModelListener(tableChangeListener);
    }
    
    /**
     * Panel which holds all buttons and their actions for tables.
     *  
     * @author rjwaal
     * @since 1.3
     * @version DC1.00
     */    
    public class TableActionsPanel extends JPanel {
        private JButton buttonClear;
        private JButton buttonSave;
        private JButton buttonAddRow;
        private JButton buttonRemoveRow;

        public TableActionsPanel(boolean showAddRemoveRow) {
            setLayout(Layout.getGBL());

            // Create the components
            buttonSave = ComponentFactory.getButton(Resources.getText(DCLabels.lblSaveChanges));
            buttonClear = ComponentFactory.getButton(Resources.getText(DCLabels.lblClear));
            buttonAddRow = ComponentFactory.getButton(Resources.getText(DCLabels.lblAddRow));
            buttonRemoveRow = ComponentFactory.getButton(Resources.getText(DCLabels.lblRemoveRow));

            buttonSave.setMnemonic(KeyEvent.VK_S);
            buttonClear.setMnemonic(KeyEvent.VK_C);
            buttonAddRow.setMnemonic(KeyEvent.VK_A);
            buttonRemoveRow.setMnemonic(KeyEvent.VK_R);

            buttonSave.setToolTipText(Resources.getText(DCTooltips.tpSaveChanges));
            buttonClear.setToolTipText(Resources.getText(DCTooltips.tpClear));
            buttonAddRow.setToolTipText(Resources.getText(DCTooltips.tpAddRow));
            buttonRemoveRow.setToolTipText(Resources.getText(DCTooltips.tpRemoveRow));

            buttonClear.addActionListener(new ClearButtonAction());
            buttonAddRow.addActionListener(new AddRowAction());
            buttonRemoveRow.addActionListener(new RemoveRowAction());

            // Build the panel
            if (showAddRemoveRow) {
                add(buttonAddRow,   Layout.getGBC( 0, 0, 1, 1, 1.0, 1.0
                                   ,GridBagConstraints.SOUTHWEST, GridBagConstraints.NONE,
                                    new Insets(5, 5, 5, 5), 0, 0));
                add(buttonRemoveRow,Layout.getGBC( 1, 0, 1, 1, 1.0, 1.0
                                   ,GridBagConstraints.SOUTHWEST, GridBagConstraints.NONE,
                                    new Insets(5, 5, 5, 5), 0, 0));
            }

            add(buttonSave,     Layout.getGBC( 2, 0, 1, 1, 1.0, 1.0
                               ,GridBagConstraints.SOUTHWEST, GridBagConstraints.NONE,
                                new Insets(5, 5, 5, 5), 0, 0));

            add(buttonClear,    Layout.getGBC( 3, 0, 1, 1, 1.0, 1.0
                               ,GridBagConstraints.SOUTHWEST, GridBagConstraints.NONE,
                                new Insets(5, 5, 5, 5), 0, 0));
        }

        /** 
         * Allows additional action listener to be added to the 
         * save button. Sequence of execution of these actions cannot
         * be garuanteed.
         * 
         * @param listener action listener to be added
         */
        public void addSaveActionListener(ActionListener listener) {
            buttonSave.addActionListener(listener);
        }

        /** 
         * Removes action listeners from the save button. 
         * 
         * @param listener action listener to be removed 
         */        
        public void removeSaveActionListener(ActionListener listener) {
            buttonSave.removeActionListener(listener);
        }
    }

    /**
     * Opens a record
     */
    public class OpenRecordAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            openRecord();
        }
    }

    /**
     * Clears the table
     */
    public class ClearButtonAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            clearTable();
        }
    }

    /**
     * Handles changes made to the table (on screen)
     */
    private class TableValueChangedAction implements TableModelListener {
        public void tableChanged(TableModelEvent e) {
            if (    e.getType() == TableModelEvent.INSERT || 
                    e.getType() == TableModelEvent.UPDATE) {
                int row = table.getSelectedRow();
                addRowToCache(row, e.getColumn());
            }
        }
    }
    
    /**
     * Opens the hidden / visible fields selector dialog 
     */
    public class FieldSelectorAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            startFieldSelector();
        }        
    } 

    /**
     * Opens the dialog for updating all the values of one (or more)
     * columns in the UI table
     */
    public class UpdateAllAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            UpdateAllDialog dialog =  new UpdateAllDialog(getModule().getDataCrowObject(),
                                                          DataCrow.mainFrame);
            try {
                // if the update was a success
                if (dialog.isSuccess()) {
                    dialog.setModal(false);
                    // retrieve the data which will be applied on the table
                    Object[] values = dialog.getData();
                    
                    int rows = tableModel.getRowCount();
                    
                    tableModel.removeTableModelListener(tableChangeListener);
                    
                    for (int i = 0; i < rows; i++) {
                        for (int j = 0; j < values.length; j++) {
                            // only update a field with a 'real' / not empty value
                            if (values[j] != null &&
                                !StringUtils.isEmpty(values[j].toString()) &&
                                !values[j].toString().equals("-1")) {
                                tableModel.setValueAt(values[j], i, j);
                                addRowToCache(i, j);
                            }
                        } 
                    }
                    tableModel.addTableModelListener(tableChangeListener);
                }
            } catch (Exception exp) {
                Messages.setMessage(exp);
            } finally {
                dialog.dispose();
                dialog = null;
            }
        }        
    }    

    /**
     * Adds an empty row to the table
     */
    public class AddRowAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            LinkedList row = new LinkedList();
            row.add(getModule().getDataCrowObject());
            addRow(row);
        }
    }

    /**
     * Removes a row (or multiple rows) from the table
     */
    public class RemoveRowAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            if (tableModel.getRowCount() < 1 ) {
                return;
            }
            int[] rows = table.getSelectedRows();
            if (rows.length == 0) {
                tableModel.removeRow(tableModel.getRowCount() -1);

            } else {
                for (int i = rows.length; i > 0; i--) {
                    tableModel.removeRow(rows[i-1]);
                }
            }
        }
    }

    /**
     * Listens on right clicks:
     * shows the popup menu 
     * 
     * Listens on double clicks:
     * opens the record form for the row on which is double clicked
     */
    private class TableMouseListener implements MouseListener {
        public void mouseReleased(MouseEvent e) {
            if (SwingUtilities.isRightMouseButton(e)) {
                if (table.getSelectedRows() != null &&
                    table.getSelectedRows().length < 2) {
                    int row = table.rowAtPoint(e.getPoint());
                    int col = table.getSelectedColumn() != -1 ? table.getSelectedColumn() :
                                                                table.columnAtPoint(e.getPoint());
                    table.addRowSelectionInterval(row, row);
                    table.addColumnSelectionInterval(col, col);
                    
                    ((DCTableCellRenderer) table.getCellRenderer(row, col)).
                        getTableCellRendererComponent(
                                table, 
                                tableModel.getValueAt(row, col),
                                true,
                                true,
                                row,
                                col);
                    
                }
                popupMenu.show(table, e.getX(), e.getY());
            }

            if (e.getClickCount() == 2) {
                openRecord();
            }
        }
        public void mouseEntered(MouseEvent e) {}
        public void mouseExited(MouseEvent e) {}
        public void mousePressed(MouseEvent e) {}
        public void mouseClicked(MouseEvent e) {}
    }
    
    /**
     * Opens the PDF creation dialog 
     */
    public class CreatePdfReport implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            createPdfReport();
        }        
    }
}
