/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 Robert-Jan van der Waals                    *
 *                             rwaals@anytimenow.com                          *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.console;

import java.awt.Color;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.util.Enumeration;
import java.util.Vector;

import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JViewport;
import javax.swing.ToolTipManager;

import net.sf.dc.console.components.ComponentFactory;
import net.sf.dc.console.components.Layout;
import net.sf.dc.console.software.SoftwareCategoryForm;
import net.sf.dc.core.IconLibrary;
import net.sf.dc.core.objects.DCCollection;
import net.sf.dc.core.objects.DCObject;
import net.sf.dc.core.objects.Requestor;
import net.sf.dc.core.objects.RequestorCollection;
import net.sf.dc.db.QueryOptions;
import net.sf.dc.resources.DCLabels;
import net.sf.dc.resources.DCMessages;
import net.sf.dc.resources.DCTooltips;
import net.sf.ext.util.Resources;

/**
 * Base for all search panels (for objects like Software, Mp3 and Movie)
 *
 * @author Robert-Jan van der Waals
 * @since 1.3
 * @version DC0.50
 */
public abstract class TableSearchPanel extends TablePanel {

    protected SearchOptionsPanel panelSearchOpts;
    protected SearchConditionsPanel panelSearchConds;
    protected JPanel panelStatus;
    protected JPanel panelResult;
    protected JPanel panelSearchActs;
    protected JScrollPane scrollSearchConds;
    
    protected JButton buttonSearch;
    
    /** 
     * 
     */
    public SearchOptionsPanel getSearchOptionPanel() {
        return panelSearchOpts;
    }

    public JScrollPane getSearchConditionsPanel() {
        return scrollSearchConds;
    }
    
    public JPanel getSearchActionsPanel() {
        return panelSearchActs;
    }
    
    public JPanel getStatusPanel() {
        return panelStatus;
    }

    /**
     * Initializes the panel. Should be called by the child by calling
     * super(module). The popup menu and the action panel are automatically
     * created.
     *
     * @param module the module indentifier
     */
    public TableSearchPanel(int module) {
        super(module);

        boolean imdb = getModule().isImdbUpdateAllowed();
        boolean amazon = getModule().isAmazonUpdateAllowed();
        boolean open = this instanceof SoftwareCategoryForm.CatagorySearchPanel ?
                       false : true;

        panelSearchOpts = new SearchOptionsPanel();
        panelSearchConds = new SearchConditionsPanel();

        panelTableActions = new TableActionsPanel(false);
        panelTableActions.addSaveActionListener(new SaveUpdatedValuesAction());
    
        popupMenu = new PopupMenu(imdb, amazon, true, false, open, true);
        popupMenu.addAmazonActionListener(new AmazonUpdateAction());
        popupMenu.addIMDBActionListener(new ImdbUpdateAction());
        popupMenu.addSaveActionListener(new SaveSelectedValuesAction());
        popupMenu.addDeleteActionListener(new DeleteValuesAction());
        
        buildPanel();
        
        for (Enumeration enum = table.getColumnModel().getColumns();
             enum.hasMoreElements(); ) {
            
            vColumns.addElement(enum.nextElement());
        }
    }

    /**
     * Saves the selected values
     */
    public void saveSelectedValues() {
        cancelEdit();
        int[] rows = table.getSelectedRows();
        DCCollection collection = new DCCollection();
        if (rows != null && rows.length > 0) {
            for (int i = 0; i < rows.length; i++) {
                
                String sID = tableModel.getValueAt(rows[i], 0).toString();
                
                DCObject obj = (DCObject) cache.get(sID);
                
                if (obj != null) {
                    Object[] oldVals = obj.getValues();
                    Object[] newVals = getValuesAtRow(rows[i], true);
    
                    for (int j = 0; j < newVals.length; j++) {
                        if (!oldVals[j].equals(newVals[j])) {
                            obj.setValue(j, newVals[j]);
                        }
                    }                
                    collection.addObject(obj, new Integer(rows[i]));
                } else {
                	new MessageBox(Resources.getText(DCMessages.msgNoChangesToSave), MessageBox._INFORMATION);   
                }
            }
        }

        collection.saveUpdate(new RequestorCollection());
        setStatus(Resources.getText(DCMessages.msgSaveHasBeenExecuted));
    }

    /**
     * Saves all the changed values. The orginal values are stored 
     * in the 'cache' hashtable.
     */
    public void saveChangedValues() {
        cancelEdit();
        
        DCCollection collection = new DCCollection();

        for (Enumeration enum = cache.keys(); enum.hasMoreElements(); ) {   
            String sID = enum.nextElement().toString();
            DCObject obj = (DCObject) cache.get(sID);
            int row = getRowNumberWithID(sID);
            collection.addObject(obj, new Integer(row));
        }

        cache.clear();
        collection.saveUpdate(new RequestorCollection());
        setStatus(Resources.getText(DCMessages.msgSaveHasBeenExecuted));
    }

    /**
     * Searches for the Software objects by using the conditions and options
     * as specified.
     */
    public void searchValues() {
        DCObject o = panelSearchConds.getSearchValues();
        tableModel.removeTableModelListener(tableChangeListener);

        Requestor requestor = new Requestor(this.getClass(), "fillTable", null);
        RequestorCollection requestors = new RequestorCollection();
        requestors.addRequestor(requestor);

        o.load(panelSearchOpts.getQueryOptions(), requestors);
        setStatus(Resources.getText(DCMessages.msgSearchHasBeenExecuted));

        tableModel.addTableModelListener(tableChangeListener);
        cache.clear();
    }
    
    /**
     * Permanently removes a record / records from the database.
     * The user must always confirm the delete action first before 
     * it is executed !
     */
    public void deleteValues() {
        cancelEdit();
        int[] rows = table.getSelectedRows();

        if (rows == null || rows.length <= 0) {
            new MessageBox(Resources.getText(DCMessages.msgNoRowsSelected),
                           MessageBox._WARNING);
            return;
        }

        DCCollection collection = new DCCollection();

        for (int i = 0; i < rows.length; i++) {
            DCObject o = getModule().getDataCrowObject();
            o.setValue(DCObject._ID, tableModel.getValueAt(rows[i], 0));
            collection.addObject(o, new Integer(rows[i]));
        }

        new QuestionBox(Resources.getText(DCMessages.msgDeleteQuestion));

        if (QuestionBox.result.booleanValue()) {
            RequestorCollection requestors = new RequestorCollection();
            Requestor request = new Requestor(getClass(), "removeRows",
                                              new Object[] {table.getSelectedRows()});
            requestors.addRequestor(request);
            collection.delete(requestors);

            setStatus(Resources.getText(DCMessages.msgDeleteHasBeenExecuted));
        }
    }
    
    /**
     * Removes hidden fields from the search table.
     */
    public void removeHiddenFields(int[] indices) {
        panelSearchConds.removeHiddenFields(indices);
        super.removeHiddenFields(indices);  
    }
    
    /**
     * Creates a panel for creating the search condition.
     * The search condition exists out of fields with a value.
     * On this combination will be searched... Hidden fields are 
     * not shown in this panel.
     * 
     * @author Nuclear Vision
     */
    public class SearchConditionsPanel extends JPanel {
        private final DCObject o;
        
        private final Vector vInput = new Vector(0);
        private final Vector vLabels = new Vector(0);
        
        public SearchConditionsPanel() {
            o = getModule().getDataCrowObject();
            buildPanel(); 
            removeHiddenFields(o.getHiddenFields());
        }
        
        /**
         * Creates the panel
         */
        private void buildPanel() {
            String[] labels = o.getLabels();
            int[] fields = o.getFieldIndices();
            int[] maxLengths = o.getMaxFieldLengths();
            int[] fieldTypes = o.getFieldTypes();

            setLayout(Layout.getGBL());
            int y = 0;
            for (int i = 0; i < fields.length; i++) {
                int field = fields[i];
                JLabel label = ComponentFactory.getLabel(labels[field]);
                
                JComponent input = null;
                if (fieldTypes[field] == ComponentFactory._LONGTEXTFIELD) {
                    input = ComponentFactory.getComponent(ComponentFactory._SHORTTEXTFIELD, 
                                                          labels[field],
                                                          maxLengths[field]); 
                } else {
                    input = ComponentFactory.getComponent(fieldTypes[field], 
                                                          labels[field],
                                                          maxLengths[field]); 
                }

                int bottom = 0;
                if (i == fields.length -1) {
                    bottom = 0;   
                }
    
                add( label, Layout.getGBC( 0, y, 1, 1, 1.0, 1.0
                           ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                            new Insets(0,  5, bottom, 5), 0, 0));
                add( input, Layout.getGBC( 1, y, 1, 1, 5.0, 5.0
                           ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                            new Insets(0, 5, bottom,  5), 0, 0));

                vLabels.addElement(label);
                vInput.addElement(input);
                y++;
            }
        }

        /**
         * Makes sure that hidden fields are not displayed
         * @param indices hidden field indices
         */
        public void removeHiddenFields(int[] indices) {
            int[] notSearchable = o.getNotSearchableFields();

            // first, set all fields to be visible
            for (int i = 0; i < vInput.size(); i++) {
                JComponent comp = (JComponent) vInput.elementAt(i);
                comp.setVisible(true);
                JLabel label = (JLabel) vLabels.elementAt(i);
                label.setVisible(true);                
            }

            // secondly, hide hidden fields
            for (int i = 0; i < indices.length; i++) {
                JComponent comp = (JComponent) vInput.elementAt(indices[i]);
                comp.setVisible(false);
                JLabel label = (JLabel) vLabels.elementAt(indices[i]);
                label.setVisible(false);
            }
            
            // removes fields on which cannot be searched (such as picture fields)
            for (int i = 0; i < notSearchable.length; i++) {
                JComponent comp = (JComponent) vInput.elementAt(notSearchable[i]);
                comp.setVisible(false);
                JLabel label = (JLabel) vLabels.elementAt(notSearchable[i]);
                label.setVisible(false);
            }
        }
                
        /**
         * Returns the values on which will be searched (the search condition)
         */
        public DCObject getSearchValues() {
            DCObject obj = getModule().getDataCrowObject();
            
            final Object[] values = obj.getValues();
            int[] visibleFields = obj.getVisibleFields();
            int[] fieldTypes = obj.getFieldTypes();

            for (int i = 0; i < visibleFields.length; i++) {
                int fieldType = fieldTypes[visibleFields[i]];
                JComponent component = (JComponent) vInput.elementAt(visibleFields[i]);
                
                if (fieldType == ComponentFactory._LONGTEXTFIELD) {
                    values[visibleFields[i]] = ComponentFactory.getComponentValue(
                                                    ComponentFactory._SHORTTEXTFIELD,
                                                    component);
                } else {
                    values[visibleFields[i]] = ComponentFactory.getComponentValue(
                                                    fieldType,
                                                    component);    
                }
            }
            
            for (int i = 0; i < values.length; i++) {
                String s = values[i] == null ? "" : values[i].toString();
                if (s != null && !s.equals("")) {
                    obj.setValue(i, values[i]);
                }
            }
            return obj;
        }
    }

    /**
     * Search Option panel. Here options, such as 'comply to all conditions', 
     * are maintained
     * @author Nuclear Vision
     */
    public class SearchOptionsPanel extends JPanel {
        private JCheckBox checkComplyAll = 
            ComponentFactory.getCheckBox(Resources.getText(DCLabels.lblAndOrSelect));
        private JCheckBox checkPrecise =
            ComponentFactory.getCheckBox(Resources.getText(DCLabels.lblPreciseSelect));
        private JLabel labelOrderBy = ComponentFactory.getLabel(
                                      Resources.getText(DCLabels.lblOrderBy));
        private JComboBox comboOrderBy = ComponentFactory.getComboBox();

        public SearchOptionsPanel() {
            DCObject o = getModule().getDataCrowObject();
            setLayout(Layout.getGBL());

            String[] colNames    = o.getLabels();
            comboOrderBy.addItem("");
            for (int i = 0; i < colNames.length; i++) {
                comboOrderBy.addItem(colNames[i]);
            }
            
            comboOrderBy.setSelectedItem(colNames[1]);
            checkComplyAll.setSelected(true);
            
            // Build the panel
            add( checkComplyAll, Layout.getGBC( 0, 0, 2, 1, 1.0, 1.0
                                ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                                 new Insets( 0, 5, 3, 5), 0, 0));
            add( checkPrecise,   Layout.getGBC( 0, 1, 2, 1, 5.0, 5.0
                                ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                                 new Insets( 0, 5, 3, 5), 0, 0));
            add( labelOrderBy,   Layout.getGBC( 0, 2, 1, 1, 1.0, 1.0
                                ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                                 new Insets( 0, 5, 0, 5), 0, 0));
            add( comboOrderBy,   Layout.getGBC( 1, 2, 1, 1, 5.0, 5.0
                                ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                                 new Insets( 0, 5, 5, 5), 0, 0));
        }

        /**
         * Returns the query options used for searching values
         *
         * @return QueryOptions
         */
        public QueryOptions getQueryOptions() {
            boolean bPreciseSelect = checkPrecise.isSelected();
            boolean bComplyToAllCondition = checkComplyAll.isSelected();
            String sOrderBy = "";
            if (comboOrderBy.getSelectedIndex() > 0) {
                sOrderBy = getModule().getFieldDefinition(comboOrderBy.getSelectedIndex() - 1).getDatabaseFieldName();
            }
            return new QueryOptions(sOrderBy, bComplyToAllCondition, bPreciseSelect);
        }
    }

    /**
     * Initializes the panel
     */
    private void buildPanel() {
        panelStatus     = new JPanel(false);
        panelResult     = new JPanel(false);
        panelSearchActs = new JPanel(false);

        // ************************************************************************
        // SEARCH ACTION PANEL

        buttonSearch = ComponentFactory.getButton(Resources.getText(DCLabels.lblSearch));
        
        buttonSearch.setIcon(IconLibrary._icoFetch);
        
        buttonSearch.setToolTipText(Resources.getText(DCTooltips.tpSearch));
        buttonSearch.addActionListener(new SearchButtonAction());
        buttonSearch.setMnemonic(KeyEvent.VK_F);
        
        buttonSearch.requestFocus();
 
        panelSearchActs.add( buttonSearch,       Layout.getGBC( 0, 0, 1, 1, 1.0, 1.0
                            ,GridBagConstraints.NORTHWEST, GridBagConstraints.NONE,
                             new Insets( 0, 5, 0, 5), 0, 0));

        // ************************************************************************
        // RESULT PANEL
        // Set layout style
        panelResult.setLayout(Layout.getGBL());

        table = ComponentFactory.getTableForObject(getModule().getDataCrowObject(), 
                                                   tableModel);
        tableModel.addTableModelListener(tableChangeListener);
        table.addMouseListener(tableMouseListener);

        JScrollPane scrollResults = new JScrollPane(table);
        scrollResults.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
        scrollResults.getViewport().setScrollMode(JViewport.SIMPLE_SCROLL_MODE);
        
        // Build the panel
        panelResult.add(    scrollResults,  Layout.getGBC( 0, 0, 1, 1, 80.0, 80.0
                           ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
                            new Insets(5, 5, 5, 5), 0, 0));

        // ************************************************************************
        // STATUS PANEL
        // Set layout style
        panelStatus.setLayout(Layout.getGBL());

        // Initialize components
        labelStatus = ComponentFactory.getLabel(Resources.getText(DCLabels.lblVersion));

        // Build the panel
        panelStatus.add(    labelStatus,       Layout.getGBC( 0, 0, 1, 1, 1.0, 1.0
                           ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                            new Insets(0, 0, 0, 0), 0, 0));


        // ************************************************************************
        // MAIN PANEL
        // Set layout style
        setLayout(Layout.getGBL());

        // Initialize components
        panelResult.setBorder(ComponentFactory.getTitleBorder(
                              Resources.getText(DCLabels.lblSearchResult)));
        panelStatus.setBorder(ComponentFactory.getTitleBorder(
                              Resources.getText(DCLabels.lblStatus)));
        panelSearchOpts.setBorder(ComponentFactory.getTitleBorder(
                                  Resources.getText(DCLabels.lblSearchOpts)));

        scrollSearchConds = new JScrollPane(panelSearchConds);
        scrollSearchConds.setBorder(ComponentFactory.getTitleBorder(
                         Resources.getText(DCLabels.lblSearchCond)));

        scrollSearchConds.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        scrollSearchConds.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED);
        
        // Build the panel
        add(    scrollSearchConds,Layout.getGBC( 0, 0, 1, 2, 20.0, 20.0
               ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
                new Insets(5, 5, 5, 5), 0, 0));
        add(    panelSearchOpts,  Layout.getGBC( 1, 0, 1, 1, 1.0, 1.0
               ,GridBagConstraints.NORTHEAST, GridBagConstraints.HORIZONTAL,
                new Insets(5, 5, 5, 5), 0, 0));
        add(    panelSearchActs,  Layout.getGBC( 1, 1, 1, 1, 1.0, 1.0
               ,GridBagConstraints.SOUTHWEST, GridBagConstraints.NONE,
                new Insets(5, 5, 5, 5), 0, 0));
        add(    panelResult,      Layout.getGBC( 0, 3, 3, 1, 100.0, 100.0
               ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
                new Insets(5, 5, 5, 5), 0, 0));
        add(    panelTableActions,Layout.getGBC( 0, 4, 3, 1, 1.0, 1.0
               ,GridBagConstraints.NORTHWEST, GridBagConstraints.NONE,
                new Insets(5, 5, 5, 5), 0, 0));
        add(    panelStatus,      Layout.getGBC( 0, 5, 3, 1, 1.0, 1.0
               ,GridBagConstraints.SOUTH, GridBagConstraints.HORIZONTAL,
                new Insets(5, 5, 5, 5), 0, 0));

        ToolTipManager.sharedInstance().registerComponent(table);
    }

    public void setBackground(Color color) {
        super.setBackground(color);
        
        if (panelSearchOpts != null) {
            panelSearchOpts.setBackground(color);
            panelSearchOpts.checkPrecise.setBackground(color);
            panelSearchOpts.checkComplyAll.setBackground(color);
        }
        
        if (scrollSearchConds != null) {
            scrollSearchConds.setBackground(color);
            panelSearchConds.setBackground(color);
        }
    }

    // ****************************************************************************
    // ACTIONS
    // ****************************************************************************
    public class SearchButtonAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            cancelEdit();
            searchValues();
        }
    }

    private class SaveSelectedValuesAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            saveSelectedValues();
        }
    }

    private class AmazonUpdateAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            startAmazonUpdate();
        }
    }

    private class SaveUpdatedValuesAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            saveChangedValues();
        }
    }

    private class DeleteValuesAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            deleteValues();
        }
    }

    private class ImdbUpdateAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            cancelEdit();
            startIMDBUpdate();
        }
    }
}
