/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 Robert-Jan van der Waals                    *
 *                             rwaals@anytimenow.com                          *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.console.components;

import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Insets;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.net.URL;

import javax.swing.BorderFactory;
import javax.swing.ImageIcon;
import javax.swing.JComponent;

import net.sf.dc.console.PictureView;
import net.sf.dc.core.IconLibrary;
import net.sf.dc.messages.Messages;
import net.sf.dc.resources.DCMessages;
import net.sf.dc.resources.DCTooltips;
import net.sf.ext.jflexsett.settings.ui.ISettingsComponent;
import net.sf.ext.util.Resources;
import net.sf.ext.util.Utilities;

/**
 * Shows a scaled (or not scaled) picture. Also has a zoom function.
 * On a double click the picture view is opened.
 * 
 * @see net.sf.dc.console.PictureView
 * 
 * @author Robert-Jan van der Waals
 * @since 1.3
 * @version DC1.00
 */
public class PictureField extends JComponent implements ISettingsComponent {

    // url to the cached image
    private URL url = null;
    
    private boolean scaled = true;
    private Dimension size = null;
    
    private Dimension iconSize = null;
    private Insets insets = new Insets(0, 0, 0, 0);

    /**
     * Creates this field
     * 
     * @param allowZoom Indicates if the picture view can be opened
     */
    public PictureField(boolean allowZoom) {
        this(true, allowZoom);
    }

    /**
     * Creates this field
     * 
     * @param allowZoom Indicates if the picture view can be opened
     * @param scaled Indicates whether the picture is allowed to be
     *        scaled
     */
    public PictureField(boolean scaled, boolean allowZoom) {
        this.scaled = scaled;
        
        if (allowZoom) {
            this.setToolTipText(Resources.getText(DCTooltips.tpPictureField));
            addMouseListener(new OpenPictureFrame());
        }
        super.setBorder(BorderFactory.createRaisedBevelBorder()); 
    }

    public void setScaleImage(boolean b) {
        scaled = b;
    }
    
    /**
     * Sets a calue on this field
     * 
     * @param o path / url
     */
    public void setValue(Object o) {
        if (o == null) {
            url = null;
        } else if (o instanceof URL) {
        	url = (URL) o;
        } else {
        	String s = o.toString();
            if (s.endsWith("jpg")) {
                try {
                    if (!s.startsWith("file:")) {
                    	url = new URL("file:" + s);
                    } else {
                        url = new URL(s);
                    }
                } catch (Exception exp) {
                    Messages.setMessage(
                            Resources.getText(DCMessages.msgCouldNotLoadPicture),
                            Messages._ERROR);
                	Messages.setMessage(exp);
                }
            } else if (s.length() > 1) {
                url = Utilities.createCachedImage(s);
            }
            s = null;
            o = null;
        }
        revalidate();
    }
    
    /**
     * Returns the image as a BASE64 string
     */
    public Object getValue() {
        return Utilities.fileToBase64String(url);
    }
    
    /**
     * Returns the java.awt.Image
     */
    private Image getImage() {
        try {
            if (url != null) {
                ImageIcon icon = new ImageIcon(url);
                iconSize = new Dimension(icon.getIconWidth(), icon.getIconHeight());
                return icon.getImage();
            } else {
                return null;
            }
        } catch (Exception exp) {
            Messages.setMessage(exp);
            return null;
        }        
    }
    
    /**
     * Indicates if scaling is allowed. Scaling is not allowed for very 
     * small pictures as otherwise they will just become disrupted..
     */
    private boolean scalingAllowed() {
        if (scaled && (iconSize.height >= 50 && iconSize.width >= 50)) {
        	return true;
        } else {
        	return false;
        }
    }
    
    /**
     * Paints the image. This is either done in a scaled way 
     * (@see #scalingAllowed(Dimension)) or in it is orginal state
     * 
     */
    public void paint(Graphics g) {
        insets = getInsets(insets);
        size = getSize(size);
        
        Image image = getImage();
        
        if (image != null && scalingAllowed()) {
            int width = size.width - insets.left - insets.right;
            int height = size.height - insets.top - insets.bottom;
            g.drawImage(image, insets.left, insets.top, width, height, this);
            image.flush();
        } else if (image != null){
            g.drawImage(image, insets.left, insets.top, this);
            image.flush();
        } else {
            g.drawImage(IconLibrary._icoNoPic, insets.left, insets.top,
                        size.width - insets.left - insets.right,
                        size.height - insets.top - insets.bottom,
                        this);
        }
        g.dispose();
    }

    /**
     * Calculates the minimum size for this component
     */
    public Dimension getMinimumSize() {
        int imgw = 32, imgh = 32;
        Image image = getImage();
        if (image != null) {
            imgw = image.getWidth(this);
            imgh = image.getHeight(this);
            image.flush();
        }
        insets = getInsets(insets);
        return new Dimension(
            insets.left + Math.max(32, imgw / 10) + insets.right,
            insets.top + Math.max(32, imgh / 10) + insets.bottom);
    }

    /**
     * Calculates the preferred size for this component
     */
    public Dimension getPreferredSize() {
        int imgw = 32, imgh = 32;
        Image image = getImage();
        if (image != null) {
            imgw = image.getWidth(this);
            imgh = image.getHeight(this);
            image.flush();
        }
        insets = getInsets(insets);
        return new Dimension(
            insets.left + imgw + insets.right,
            insets.top + imgh + insets.bottom);
    }

    /**
     * Calculates the maximum size for this component
     */
    public Dimension getMaximumSize() {
        return getPreferredSize();
    }

    /**
     * Opens the picture viewer (@see PictureView)
     */
    public void showFullPicture() {
        PictureView picture = new PictureView(url);
        
        if (!picture.wasClosed()) {
        	setValue(picture.getPictureBase64());
        }
        picture = null;
    }
    
    public class OpenPictureFrame implements MouseListener {
        public void mouseReleased(MouseEvent e) {
            if (e.getClickCount() == 2) {
                showFullPicture();
            }
        }
        public void mouseEntered(MouseEvent e) {}
        public void mouseExited(MouseEvent e) {}
        public void mousePressed(MouseEvent e) {}
        public void mouseClicked(MouseEvent e) {}
    }  
}