/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 Robert-Jan van der Waals                    *
 *                             rwaals@anytimenow.com                          *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.console.movie;

import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import net.sf.dc.console.MessageBox;
import net.sf.dc.console.SearchForm;
import net.sf.dc.core.Repository;
import net.sf.dc.core.modules.DCModules;
import net.sf.dc.core.objects.movie.Movie;
import net.sf.dc.http.IMDBMovieQuery;
import net.sf.dc.messages.Messages;
import net.sf.dc.resources.DCMessages;
import net.sf.ext.jflexsett.settings.Settings;
import net.sf.ext.util.Resources;
import net.sf.ext.util.SwingUtil;

/**
 * Form used to search for IMDB data. The search is performed on the
 * title of the movie
 *
 * @see net.sf.dc.console.SearchForm
 * @author Robert-Jan van der Waals
 * @since 1.3
 * @version DC0.50
 */
public class ImdbMovieSearchForm extends SearchForm {
    
    /**
     * Instantiates this form and starts the search
     * 
     * @param rowNumber the rownumber of the to be updated record
     * @param title the title for this form
     * @param name the name on which to search
     * @param sender the requestor to which feedback should be given 
     */
    public ImdbMovieSearchForm(int rowNumber, 
							   String title, 
							   String name, 
							   Object sender) {
                              
        super(sender, new int[] {Movie._A_TITLE,
                                 Movie._B_DESCRIPTION,
                                 Movie._C_WEBPAGE,
                                 Movie._D_GENRE,
                                 Movie._E_ACTORS,
                                 Movie._F_DIRECTOR,
                                 Movie._G_COUNTRY,
                                 Movie._H_YEAR,
                                 Movie._I_PLAYLENGTH,
                                 Movie._U_PICTUREFRONT},
               DCModules._MOVIE);
        super.setTitle(title);
        SwingUtil.setRootFrame(this);
        
        isUpdateRunning = true;
        addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                Settings.setValue(Repository.settings.stImdbSearchFormViewSize,
                                  getSize());
                running = false;
            }
        });        
        
        this.rowNumber = rowNumber;
        
        textSearch.setText(name);
        startSearch();
        
        setVisible(true);
    }

    /**
     * Starts the search thread
     * @see SearchJob
     */
    protected void startSearch() {
        String name = textSearch.getText();
        
        if (name == null || name.equals("")) {
            return;   
        }
                
        threadSearch = new Thread(new SearchJob(name), "SearchJob");
        threadSearch.setDaemon(true);
        threadSearch.setPriority(1);
        threadSearch.start();        
    }
    
    /**
     * Simple search thread. Uses a query object to contact the
     * internet and puts the results in the table of this form
     */
    private class SearchJob extends Thread {
        
        String name = "";
        
        public SearchJob(String name) {
            this.name = name;
        }
        
        public void run() {
            try {
                running = true;
                buttonSearch.setEnabled(false);
                buttonStop.setEnabled(true);
                buttonStopAll.setEnabled(true);
                searchImdb(name);
            } catch (Exception exp) {
                exp.printStackTrace();
                Messages.setMessage(exp);
                new MessageBox(Resources.getText(DCMessages.msgProcessCrash), 
                               MessageBox._ERROR);
            }
        }      

        private void searchImdb(String name) {
            try {
                IMDBMovieQuery imdbQuery = new IMDBMovieQuery(name);
                String[] ids = imdbQuery.getIDs();
                if (ids == null) {
                    return;
                }
                
                progress.setMaximum(ids.length);
                progress.setValue(0);
                
                for (int i = 0; i < ids.length; i++) {
                    if (!running) {
                        return;   
                    }
                    Movie movie = imdbQuery.parseInformation(ids[i]);
                    progress.setValue(i + 1);
                    setData(movie);
                }
            } catch (Exception exp) {
                Messages.setMessage(exp);
                new MessageBox(exp.getMessage(), MessageBox._ERROR);   
            } finally {
                buttonSearch.setEnabled(true);
                buttonStop.setEnabled(false);
                buttonStopAll.setEnabled(false);
            }
        }
    }
}
