/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 Robert-Jan van der Waals                    *
 *                             rwaals@anytimenow.com                          *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.core.modules;

import javax.swing.ImageIcon;
import javax.swing.JButton;

import net.sf.dc.console.TableInsertPanel;
import net.sf.dc.console.TableSearchPanel;
import net.sf.dc.core.objects.DCFieldDefinition;
import net.sf.dc.core.objects.DCObject;
import net.sf.dc.processes.ImportCDProcess;

/**
 * Contains all generic information for this module. 
 * Only static data can be found in here, such as the column and
 * table information. The UI representation can also be retrieved
 * thru this class (insert and search panels)
 * 
 * Hidden field information and such, which can be changed during
 * a session is contained within the Data Crow Objects themselves.
 * 
 * Main reason for this setup is to keep the Data Crow Objects as
 * small as possible, they can be stored in memory by thousands
 * at a time..
 * 
 * @see net.sf.dc.core.modules.DCModules
 * 
 * @author Robert-Jan van der Waals
 * @since 1.3
 * @version DC1.00
 */
public abstract class DCModule {
    
    protected TableInsertPanel insertPanel;
    protected TableSearchPanel searchPanel;
    
    // all fields for this module
    protected DCFieldDefinition[] fields;
    
    private final int index;
    private final String name;
    
    private final boolean allowImdbUpdate;
    private final boolean allowAmazonUpdate;
    
    // for later use; indicates whether this module is a composite
    // module or not. For now there are no composite modules
    private final boolean standalone;
    
    protected String tableName = "";

    /**
     * Initializes this module
     * 
     * @param index static final module index
     * @param name name of this module
     * @param allowImdbUpdate internet updates, thru imdb, 
     *        allowed y/n
     * @param allowAmazonUpdate internet updates, thru amazon, 
     *        allowed y/n
     * @param standalone is this a standalone or a composite module 
     */
    public DCModule(        int index, 
                            String name, 
                            boolean allowImdbUpdate,
                            boolean allowAmazonUpdate,
                            boolean standalone) {
        
        this.index = index;
        this.name = name;
        this.allowAmazonUpdate = allowAmazonUpdate;
        this.allowImdbUpdate = allowImdbUpdate;
        this.standalone = standalone;
        initializeFields();
    }
    
    /**
     * Returns a specific field definition. A field definition
     * describes if the field is required, what datatype it represents,
     * etc.
     * 
     * @param index field index
     */
    public DCFieldDefinition getFieldDefinition(int index) {
        return fields[index];
    }

    /**
     * Returns all field definitions. A field definition
     * describes if the field is required, what datatype it represents,
     * etc.
     */
    public DCFieldDefinition[] getFieldDefinitions() {
        return fields;
    }

    // to be implemented by children
    protected abstract void initializeFields();
    public abstract ImageIcon getIcon();

    /**
     * Returns an empty Data Crow object, managed by this module 
     */
    public abstract DCObject getDataCrowObject();
    
    /**
     * Returns module specific menu items  
     */
    public abstract JButton[] getMenuItems();
    
    public abstract String getHiddenFieldsSettingsKey();
    public abstract String getRequiredFieldsSettingsKey();
    
    /**
     * Creates the panels for this module
     */
    protected abstract void initializeUI();
    
    /**
     * Returns the module specific CD import process
     * 
     * @param sourceDir directory which is used as source for the
     *        import
     * @param confirmEachDir indicates whether for each directory
     *        confirmation should be asked if the underlying data
     *        of this directory should be read or not.
     * @param nameOfCD cd name (used as the location name)
     * 
     * @return
     */
    public abstract ImportCDProcess getCDImportProcess(
                                                 String sourceDir,
                                                 boolean confirmEachDir,
                                                 String nameOfCD);

    /**
     * Returns the amount of fields 
     */
    public int getFieldCount() {
        return fields.length;    
    }
    
    /**
     * Initializes the insert panel (lazy)
     */
    public TableInsertPanel getInsertPanel() {
        if (insertPanel == null) {
            initializeUI();
        }
        return insertPanel;
    }
    
    public boolean isStandalone() {
        return standalone;
    }
    
    /**
     * Initializes the search panel (lazy)
     */
    public TableSearchPanel getSearchPanel() {
        if (searchPanel == null) {
            initializeUI();
        }
        return searchPanel;
    }
    
    /**
     * Indicates if this module has an Amazon update implementation
     */
    public boolean isAmazonUpdateAllowed() {
        return allowAmazonUpdate;
    }

    /**
     * Indicates if this module has an IMDB update implementation
     */
    public boolean isImdbUpdateAllowed() {
        return allowImdbUpdate;
    }
    
    /**
     * Returns the display name of this module
     */
    public String getName() {
        return name;
    }
    
    /**
     * Returns the index for this module as defined in DCModules
     * @see DCModules 
     */
    public int getIndex() {
        return index;
    }
    
    /**
     * Creates a string representation for this module
     */
    public String toString() {
        return name;
    }
}
