/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 Robert-Jan van der Waals                    *
 *                             rwaals@anytimenow.com                          *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.core.objects;

import java.util.Enumeration;
import java.util.Hashtable;

import net.sf.dc.console.MessageBox;
import net.sf.dc.core.Repository;
import net.sf.dc.db.Query;
import net.sf.dc.db.QueryBatch;
import net.sf.dc.db.QueryWhat;
import net.sf.dc.resources.DCMessages;
import net.sf.dc.wf.WorkFlow;
import net.sf.ext.jflexsett.settings.Settings;
import net.sf.ext.util.Resources;
import net.sf.ext.util.Utilities;

/**
 * A collection of Data Crow objects. Normal actions like
 * save / update and delete can be performed on a collection
 * of objects
 *
 * @author Robert-Jan van der Waals
 * @since 1.3
 * @version DC0.50
 */
public class DCCollection {
    
    private Hashtable htObjects;
    private static final WorkFlow wf = new WorkFlow();

    /**
     * Initializes a new collection
     */
    public DCCollection() {
        htObjects = new Hashtable();
    }
    
    /**
     * Adds an object to this collection. The rownumber 
     * indicates where the object is located in the UI table
     * 
     * @param o Data Crow Object to be added to the collection
     * @param rowNumber UI row number
     */
    public void addObject(DCObject o, Integer rowNumber) {
        if (o.getValue(DCObject._ID) == null || 
           (o.getValue(DCObject._ID).toString().equals("") ||
            o.getValue(DCObject._ID).toString().equalsIgnoreCase("null"))) {

            // set the ID for this new Data Crow object
            o.setValue(DCObject._ID, Utilities.getUniqueID());
        }
        htObjects.put(rowNumber, o);
    }    
    
    public void saveUpdate(RequestorCollection requestors) {
        wf.updateCollection(this, requestors);
    }
    
    public void delete(RequestorCollection requestors) {
        wf.deleteCollection(this, requestors);   
    }

    public void saveNew(RequestorCollection requestors) {
        wf.insertCollection(this, requestors);
    }
    
    /**
     * Convert this collection to a Query Batch. The values
     * added to the batch are checked before hand:
     * - if they are unique within this batch
     * - if all required fields have been entered
     * 
     * @param queryType type of query (@see Query)
     */
    public QueryBatch toQueryBatch(int queryType) {
        QueryBatch batch = new QueryBatch();
        for (Enumeration enum = htObjects.keys(); enum.hasMoreElements(); ) {
            Integer row = (Integer) enum.nextElement();
            DCObject o = (DCObject) htObjects.get(row);
            
            batch.setSilent(o.getSilence());
            
            boolean bValid = false;
            if (queryType == Query._INSERT) {
                bValid = o.checkIntegrity(false);
            } else if (queryType == Query._UPDATE) {
                bValid = o.checkIntegrity(true);
            } else if (queryType == Query._DELETE) {
                bValid = true;   
            }
            
            if (bValid && (queryType == Query._DELETE || uniqueObject(o, row))) {
                QueryWhat what = new QueryWhat(o.getDatabaseFieldNames(),
                                               o.getTableName());
                Query query = new Query(  what, 
                                          o.toQueryCondition(false),
                                          null,
                                          queryType,
                                          new RequestorCollection(),
                                          o.getValue(DCObject._ID).toString());
                if (query != null) {
                    batch.addQuery(query);
                }     
                o = null;
            } else {
                batch =  null;
                break;
            }
        }
        htObjects.clear();
        htObjects = null;
        return batch;
    }
    
    /**
     * Checks whether the data crow object is unqiue within the 
     * collection
     * 
     * @param o the object to check for completeness and uniqueness
     * @param newRow UI row index
     */
    public boolean uniqueObject(DCObject o, Integer newRow) {
        if (Settings.getValueAsBoolean(Repository.settings.stCheckUniqueness)) {
            int[] requiredFields = o.getRequiredFields();
            Integer oldRow = null;
            
            for (Enumeration enum = htObjects.keys(); enum.hasMoreElements(); ) {
                oldRow = (Integer) enum.nextElement();
                DCObject dcTemp = (DCObject) htObjects.get(oldRow);
                
                if (!dcTemp.equals(o)) {
                    
                    boolean same = true;
                    for (int k = 0; k < requiredFields.length; k++) {
                        if (!dcTemp.getValue(requiredFields[k]).equals(
                            o.getValue(requiredFields[k]))) {
                            same = false;  
                        }
                    }

                    if (same) {
                        newRow = new Integer(newRow.intValue() + 1);
                        oldRow = new Integer(oldRow.intValue() + 1);
                        
                        new MessageBox(Resources.getText(DCMessages.msgItemInQueueNotUnique, 
                                       new String[] {o.toString(),
                                                     newRow.toString(), 
                                                     dcTemp.toString(),
                                                     oldRow.toString()}), 
                                       MessageBox._ERROR); 
                        return false;
                    }
                }
            }
        }          
        return true;
    }    
}
