/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 Robert-Jan van der Waals                    *
 *                             rwaals@anytimenow.com                          *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.db;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Enumeration;
import java.util.LinkedList;

import net.sf.dc.core.modules.DCModule;
import net.sf.dc.core.modules.DCModules;
import net.sf.dc.core.objects.DCCollection;
import net.sf.dc.core.objects.DCObject;
import net.sf.dc.core.objects.RequestorCollection;
import net.sf.dc.messages.Messages;
import net.sf.dc.resources.DCMessages;
import net.sf.dc.wf.WorkFlow;
import net.sf.ext.util.Resources;
import net.sf.ext.util.Utilities;

/**
 * Holder of the Database connection and information. Starts the
 * QueryQueue and handles update, insert and select statements.
 * 
 * On the first start, the Database will be created (if the database
 * is empty; the tables do not exist). If this fails, a message is
 * shown to the user.
 * 
 * For new records, the unique ID is created by this class. This has
 * not been perfected at this moment.
 *
 * @author Robert-Jan van der Waals
 * @since 1.3
 * @version DC0.50
 */
public class Database {

    private static Connection connection;
    private QueryQueue queue;

    /**
     * Initializes the Database, creates a new database if necessary
     */
    public Database() {
        init();
    }

    /**
     * Closes the database connection
     */
    public static void close() {
        try {
            connection.close();
        } catch (Exception exp) {
            Messages.setMessage(exp);
        }
    }
    
    /**
     * Starts the query queue and initializes the database connection
     *
     * @see DatabaseConnection
     * @see QueryQueue
     */
    public void init() {
        new DatabaseConnection();
        
        if (connection == null) {
            initializeDatabase();
        }

        queue = new QueryQueue(connection);
        Thread queryQueue = new Thread(queue, "queryQueue");

        queryQueue.setPriority(Thread.NORM_PRIORITY);
        queryQueue.start();
    }

    /**
     * Inserts a collection of data crow objects into the database
     * @param collection data crow object collection
     * @param requestors requests to be executed after insert
     */
    public void insertCollection(DCCollection collection,
                                 RequestorCollection requestors) {
        QueryBatch batch = collection.toQueryBatch(Query._INSERT);
        
        if (batch != null) {
            batch.setRequestors(requestors);        
            queue.addQueryBatch(batch);
        }
    }

    /**
     * Updates a collection of data crow objects
     * @param collection data crow object collection
     * @param requestors requests to be executed after update
     */    
    public void updateCollection(DCCollection collection,
                                 RequestorCollection requestors) {
        QueryBatch batch = collection.toQueryBatch(Query._UPDATE);
        
        if (batch != null) {
            batch.setRequestors(requestors);
            queue.addQueryBatch(batch);
        }
    }
    
    /**
     * Deletes a collection of data crow objects from the database
     * @param collection data crow object collection
     * @param requestors requests to be executed after delete
     */    
    public void deleteCollection(DCCollection collection,
                                 RequestorCollection requestors) {
        QueryBatch batch = collection.toQueryBatch(Query._DELETE);
     
        if (batch != null) {
            batch.setRequestors(requestors);   
            queue.addQueryBatch(batch);
        }
    }


    /**
     * Runs a query 
     * @param sQuery query statement (string)
     * @param requestors requests to the UI
     */
    public void runQuery(String sQuery, RequestorCollection requestors) {
        Query query = new Query(null, null, null, Query._UNDEFINED, requestors, null);
        query.setQueryString(sQuery);
        queue.addQuery(query);
    }

    /**
     * Puts the query in the queue. The queue will handle the query when time
     * comes and returns the results (if there are any) to the requestor by using
     * the RequestorObject class
     * @see RequestorObject 
     * @see QueryQueue
     * @param query query object
     */
    private void putQueryInQueue(Query query) {
        queue.addQuery(query);
    }

    /**
     * Runs a query direct on the database instead of putting it in query 
     * queue. It returns an unclosed result set. 
     * @param sQuery query to be executed
     */
    public ResultSet runQueryDirectUnclosed(String sQuery) throws Exception {
        Messages.setMessage(sQuery, Messages._QUERY);
        Statement stmt = connection.createStatement();    
        return stmt.executeQuery(sQuery);
    }

    /**
     * Does not put the query in a queue but executes it immediately.
     * Errors are returned to the caller (not thrown)
     * @param sQuery  query statement
     * @return list of data crow objects (can be null or empty)
     */
    public LinkedList runQueryDirect(String sQuery) {
        LinkedList data = null;
        try {
            Statement stmt = connection.createStatement();    
            ResultSet result = stmt.executeQuery(sQuery);
            data = WorkFlow.createDataCrowObjectArray(result);
            stmt.close();
            result.close();
        } catch (SQLException sqlExp) {
            if (!sqlExp.getMessage().equals("No ResultSet was produced")) {
                Messages.setMessage(sqlExp);
            }
        }
        Messages.setMessage(sQuery, Messages._QUERY);
        return data;
    }

    /**
     * Runs a query directly, bypassing the query queue
     * @param DataCrowObject data crow object
     */
    public LinkedList runQueryDirect(DCObject o) {
        QueryWhat queryWhat = new QueryWhat(o.getDatabaseFieldNames(), o.getTableName());
        Query query = new Query(queryWhat, o.toQueryCondition(true), null, Query._SELECT, null, null);
        Messages.setMessage(query.getQueryString(), Messages._QUERY);
        return runQueryDirect(query.getQueryString());
    }

    /**
     * Does not put the query in a queue but executes it immediately.
     * Errors are returned to the caller (not thrown)
     * @param Query Query object
     */
    public LinkedList runQueryDirect(Query query) {
        return runQueryDirect(query.getQueryString());
    }

    /**
     * Searches for values. The query is put in the query queue. The results are
     * past to the caller by using the RequestorCollection
     *
     * @param o the object which is used as the condition
     * @param options query options
     * @param requestors requestors for the result of the query (UI)
     */
    public void searchValues(DCObject o, QueryOptions options, RequestorCollection requestors) {
        QueryWhat queryWhat = new QueryWhat(o.getDatabaseFieldNames(), o.getTableName());
        Query query = new Query(queryWhat, o.toQueryCondition(true), options, Query._SELECT, requestors, null);
        query.setSilence(o.getSilence());
        putQueryInQueue(query);
    }

    /**
     * Updates values. The query is put in the query queue. Actions are
     * past to the caller by using the RequestorCollection
     * @param o  the object to be updated
     * @param requestors requestors which depend on the outcome of the query
     */
    public void updateValues(DCObject o, RequestorCollection requestors) {
        String sID = o.getValue(DCObject._ID).toString();

        QueryWhat queryWhat = new QueryWhat(o.getDatabaseFieldNames(), o.getTableName());
        Query query = new Query(queryWhat, o.toQueryCondition(false), null, Query._UPDATE, requestors, sID);
        query.setSilence(o.getSilence());
        putQueryInQueue(query);
    }

    /**
     * Inserts new values. The query is put in the query queue. Actions are
     * past to the caller by using the RequestorCollection
     * @param o the object to be inserted
     * @param requestors requestors which depend on the outcome of the query
     */
    public void insertValues(DCObject o, RequestorCollection requestors) {
        String sID = Utilities.getUniqueID();
        o.setValue(DCObject._ID, sID);

        QueryWhat queryWhat = new QueryWhat(o.getDatabaseFieldNames(), o.getTableName());
        Query query = new Query(queryWhat, o.toQueryCondition(false), null, Query._INSERT, requestors, sID);
        query.setSilence(o.getSilence());
        putQueryInQueue(query);
    }

    /**
     * Deletes values from the database. The query is put in the query queue. 
     * Actions are past to the caller by using the RequestorCollection
     * @param o the object to be deleted
     * @param requestors requestors which depend on the outcome of the query
     */
    public void deleteValues(DCObject o, RequestorCollection requestors) {
        String sID = o.getValue(DCObject._ID).toString();

        QueryWhat queryWhat = new QueryWhat(o.getDatabaseFieldNames(), o.getTableName());
        Query query = new Query(queryWhat, o.toQueryCondition(false), null, Query._DELETE, requestors, sID);
        query.setSilence(o.getSilence());
        putQueryInQueue(query);
    }

    /**
     * Checks whether the values are unique (do not exist as yet in the database).
     * Updates and inserts are checked differently
     * @param o   object to be checked
     * @param bUpdateQuery  specifies whether this is an update or an insert
     */
    public boolean uniqueValues(DCObject o, boolean bUpdateQuery) {
        String sID = o.getValue(DCObject._ID).toString();

        int[]    requiredFields     = o.getRequiredFields();
        String[] conditionFields    = new String[requiredFields.length];
        Object[] conditionValues    = new Object[requiredFields.length];

        for (int i = 0; i < requiredFields.length; i++) {
            conditionFields[i] = o.getDatabaseFieldName(requiredFields[i]);
            conditionValues[i] = o.getValue(requiredFields[i]);
        }

        QueryCondition condition = new QueryCondition(conditionFields, conditionValues, true);
        QueryOptions options = new QueryOptions(null, true, true);
        QueryWhat queryWhat = new QueryWhat(o.getDatabaseFieldNames(), o.getTableName());
        Query query = new Query(queryWhat, condition, options, Query._SELECT, null, sID);

        ResultSet result = null;
        try {
            result = runQueryDirectUnclosed(query.getQueryString());
        } catch (Exception exp) {
            Messages.setMessage(exp);
        }

        if (result == null) {
            return true;
        }

        int rows = 0;
        try {
            while (result.next()) {
                String recordID = result.getString(o.getDatabaseFieldName(
                                                   DCObject._ID));
                if (!sID.equals(recordID)) {
                    rows++;
                }
            }
            result.close();
        } catch (SQLException sqlExp) {
            Messages.setMessage(sqlExp);
        }

        if (bUpdateQuery && rows >= 1) {
            return false; // Two rows or more with the same ID
        } else if (!bUpdateQuery && rows > 0) {
            return false; // One row exists already with the same ID
        }
        return true;
    }

    /**
     * Creates a connection to the database. If no database exsists, a new
     * one is created. Demo data is loaded when this is preferred. The login dialog
     * is used to get user input
     */
    private synchronized void initializeDatabase() {
        while (connection == null) {
            connection = DatabaseConnection.connect();
        }

        Statement stmt = null;
        try {
            stmt = connection.createStatement();
        } catch (Exception sqlExp) {
            return;
        }
        
        for (Enumeration enumModules = DCModules.getModules(); 
             enumModules.hasMoreElements(); ) {

            DCModule module = (DCModule) enumModules.nextElement();
            DCObject dcObj = module.getDataCrowObject(); 
            
            try {
                String testQuery = "select * from " + dcObj.getTableName();
                ResultSet result = stmt.executeQuery(testQuery);
                initializeColumns(result.getMetaData(), dcObj);
                Messages.setMessage(Resources.getText(DCMessages.msgTableFound, dcObj.getTableName()), 
                                    Messages._STANDARD);
                result.close();
            } catch (SQLException exp) {
                Messages.setMessage(Resources.getText(DCMessages.msgTableNotFound, dcObj.getTableName()), 
                                    Messages._STANDARD);
                initializeNewDatabase(dcObj);
            }
        }
        try {
            stmt.close();
        } catch (Exception ignore) {}
    }

    /**
     * Checks if all columns of a table are present. If not, the missing columns
     * are added to the table
     * @param metaData object holding the table specifics
     * @param dcObj data crow object for which the table is created
     */
    private void initializeColumns(ResultSetMetaData metaData, DCObject dcObj) 
            throws SQLException {
        String[] columns = dcObj.getDatabaseFieldNames();
        String tableName = metaData.getTableName(1);
        
        int metaCount = metaData.getColumnCount();
        if (metaCount < columns.length) {
            Messages.setMessage(Resources.getText(DCMessages.msgTableNeedsUpgrade, tableName),
                                Messages._STANDARD);
            
            String[] metaColumns = new String[metaCount];
            for (int i = 1; i < metaCount + 1; i++) {
                metaColumns[i - 1] = metaData.getColumnName(i);
            }
            
            for (int i = 0; i < columns.length; i++) {
                boolean exists = false;
                for (int j = 0; j < metaColumns.length; j++) {
                    if (columns[i].equalsIgnoreCase(metaColumns[j])) {
                        exists = true;
                    }
                }
                if (!exists) {
                    this.runQueryDirect("alter table " + 
                                        tableName + 
                                        " add column " +
                                        dcObj.getDatabaseFieldName(i) + " " +
                                        dcObj.getField(i).getDataBaseFieldType()); 
                }
            }
        }
    }

    /**
     * Creates the tables and fields on an empty database
     * @param loadDemoData  specifies if Demo data should be loaded
     * @param DB database name
     */
    private void initializeNewDatabase(DCObject dcObj) {
        QueryWhat what = new QueryWhat(dcObj.getCreateStatements(),//dcObj.getDatabaseFieldNames(), 
                                       dcObj.getTableName());
        Query query = new Query(what, null, null, Query._CREATE, null, null);

        this.runQueryDirect(query.getQueryString());
    }
}
