/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 Robert-Jan van der Waals                    *
 *                             rwaals@anytimenow.com                          *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.db;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.RandomAccessFile;
import java.sql.Connection;
import java.sql.DriverManager;
import java.util.Properties;

import net.sf.dc.console.LoginDialog;
import net.sf.dc.console.MessageBox;
import net.sf.dc.console.QuestionBox;
import net.sf.dc.core.Repository;
import net.sf.dc.messages.Messages;
import net.sf.dc.messages.MessagesWriter;
import net.sf.dc.resources.DCMessages;
import net.sf.ext.jflexsett.settings.Settings;
import net.sf.ext.util.Resources;
import net.sf.ext.util.StringUtils;

/**
 * Maintains and creates the connection to the database
 *
 * @author Robert-Jan van der Waals
 * @since 1.3
 * @version DC0.50
 */
public class DatabaseConnection {

	public static boolean bConnected = false;
    public static boolean bServerConnection = false;
    
    /**
     * Creates the connection to the database. It shows the login dialog
     * over and over again until valid data is present to connect to the
     * database 
     */
    public static Connection connect() {
    	try {
            LoginDialog dialog = new LoginDialog();
            dialog.setVisible(true);
            dialog.dispose();            
            
            Class.forName(Settings.getValueAsString(Repository.settings.stDatabaseDriver));
            String sName = Settings.getValueAsString(Repository.settings.stConnectionString);
            
            String sConnectionString = Repository.dataBase.localString + "./data/" + sName;
            if (sName.startsWith("//")) {
                // connect to a database server
                sConnectionString = Repository.dataBase.serverString + sName;
                bServerConnection = true;
            } else if (sConnectionString.indexOf("hsql") != -1) {
                // establish connection to the local database
                bServerConnection = false;
                boolean upgraded = updateHSQLDB(sName);
                if (!upgraded) {
                	return null;
                }
            }
            checkAndCorrectSettings("./data/" + sName);
            return DriverManager.getConnection(sConnectionString, "sa", "");
    	} catch (Exception exp) {
            Messages.setMessage(exp);      
            if (exp.getMessage().startsWith("socket creation error")) {
                new MessageBox(Resources.getText(DCMessages.msgServerCouldNotBeReached), Messages._ERROR);
            } else {
                new MessageBox(exp.getMessage(), Messages._ERROR);   
            }
            return null;
    	}
    }
    
    private static void checkAndCorrectSettings(String database) {
    	try {
            File file = new File(database + ".properties");
            if (file.exists()) {
                Properties properties = new Properties();
                FileInputStream fis = new FileInputStream(file);
                properties.load(fis);
                fis.close();
                
                properties.setProperty("sql.compare_in_locale", "true");
                properties.setProperty("runtime.gc_interval", "10000");
                
                FileOutputStream fos = new FileOutputStream(database + ".properties");
                properties.store(fos, "#HSQL database");
                fos.flush();
                fos.close();
            }
        } catch (Exception exp) {
        	Messages.setMessage("Could not change the settings for the database. " +
                                "The application will be fully operational. Though, the " +
                                "orderning of search results will not be conform the operating" +
                                "system localized settings", Messages._ERROR);
        }
    }
    
    /**
     * Upgrades the database. A conversion was needed for the database used
     * with Data Crow alpha and the later versions (LONGVARCHAR delivered 
     * some nasty errors). This method upgrades the database.
     * First a backup is created of the original database. 
     * 
     * @param databaseName name of the to be upgraded database
     */
    private static boolean updateHSQLDB(String databaseName) {
        try {
        	File file = new File("./data/" + databaseName + ".script");
            
            if (!file.exists()) {
                return true;
            }
            
        	RandomAccessFile raf = new RandomAccessFile(file, "r");

            long filePointer = 0;
            long fileLength  = raf.length();
            boolean upgrading = false;
            
            StringBuffer sb = new StringBuffer();
            while (filePointer < fileLength) {
                String line = raf.readLine();
                if (line.startsWith("CREATE TABLE")) {
                	sb.append(line);
                    filePointer = raf.getFilePointer();
                } else {
                	break;
                }
            }
            raf.close();

            if (sb.indexOf("LONGVARCHAR") > -1) {
                String sBackupFile = "./data/" + databaseName + ".script.bck";
                if (!upgrading) {
                	new QuestionBox(Resources.getText(DCMessages.msgUpgradeNeeded, 
                                    sBackupFile), MessagesWriter.logForm);
                    upgrading = true;
                }
                
                if (QuestionBox.result.booleanValue()) {
                    File backup = new File(sBackupFile);
                    file.renameTo(backup);          
                    
                    file = new File("./data/" + databaseName + ".script");
                    RandomAccessFile rafTarget = new RandomAccessFile(file, "rw");
                    
                    RandomAccessFile rafSource = new RandomAccessFile(backup, "r");
                    
                    filePointer = rafSource.getFilePointer();
                    boolean headerDone = false;
                    while (filePointer < fileLength) {
                        String s = rafSource.readLine();
                        if (!headerDone && s.startsWith("CREATE TABLE")) {
                        	s = StringUtils.replaceAll(s, "LONGVARCHAR", "VARCHAR");
                        } else {
                            headerDone = true;
                        }
                        
                        if (s != null) {
                            rafTarget.write(s.getBytes());
                            rafTarget.write('\n');
                        } 
                        filePointer = rafSource.getFilePointer();
                    }
                    
                    new MessageBox(Resources.getText(DCMessages.msgUpgradeSuccessFul), 
                                   MessageBox._INFORMATION);
                    rafSource.close();
                    rafTarget.close();
                } else {
                    return false;   
                }
            } 
        } catch (Exception exp) {
            new MessageBox(Resources.getText(DCMessages.msgCouldNotUpgrade, '\n' + exp.getMessage()), 
                           MessageBox._ERROR);
            return false;
        }
        return true;
    }
}
