/******************************************************************************
*                                     __                                     *
*                              <-----/@@\----->                              *
*                             <-< <  \\//  > >->                             *
*                               <-<-\ __ /->->                               *
*                               Data /  \ Crow                               *
*                                   ^    ^                                   *
*                       (c) 2003 Robert-Jan van der Waals                    *
*                             rwaals@anytimenow.com                          *
*                                                                            *
*                                                                            *
*       This library is free software; you can redistribute it and/or        *
*        modify it under the terms of the GNU Lesser General Public          *
*       License as published by the Free Software Foundation; either         *
*     version 2.1 of the License, or (at your option) any later version.     *
*                                                                            *
*      This library is distributed in the hope that it will be useful,       *
*      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
*           Lesser General Public License for more details.                  *
*                                                                            *
*     You should have received a copy of the GNU Lesser General Public       *
*    License along with this library; if not, write to the Free Software     *
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
*                                                                            *
******************************************************************************/

package net.sf.dc.export;

import java.awt.Color;
import java.io.File;
import java.io.FileOutputStream;
import java.net.URL;
import java.util.Enumeration;
import java.util.Vector;

import javax.swing.ImageIcon;

import net.sf.dc.console.BrowserDialog;
import net.sf.dc.console.MessageBox;
import net.sf.dc.console.components.ComponentFactory;
import net.sf.dc.core.objects.DCObject;
import net.sf.dc.messages.Messages;
import net.sf.dc.resources.DCLabels;
import net.sf.dc.resources.DCMessages;
import net.sf.ext.util.FileFilter;
import net.sf.ext.util.Resources;
import net.sf.ext.util.Utilities;

import com.lowagie.text.Cell;
import com.lowagie.text.Chapter;
import com.lowagie.text.Chunk;
import com.lowagie.text.Document;
import com.lowagie.text.Font;
import com.lowagie.text.FontFactory;
import com.lowagie.text.HeaderFooter;
import com.lowagie.text.Image;
import com.lowagie.text.Paragraph;
import com.lowagie.text.Phrase;
import com.lowagie.text.Table;
import com.lowagie.text.html.HtmlWriter;
import com.lowagie.text.pdf.PdfWriter;

/**
 * Creates a PDF or HTML document / report based on a set of
 * data crow objects. Hidden and empty fields are not reported on. 
 * The format and layout of the document is simple and straight
 * forward. iText is used to created the actual report
 * 
 * @author Robert-Jan van der Waals
 * @since 1.3
 * @version DC0.90
 */
public class TableToFileExporter {
    
	public static final int _PDF = 0;
	public static final int _HTML = 1;
	
	private final int reportType;
	private final DCObject[] objects;
	private final File target;

    private final Font fontMedium =FontFactory.getFont(FontFactory.COURIER, 11, 
                                                       Font.BOLD, 
                                                       new Color(0, 0, 0));
    private final Font fontSmall = FontFactory.getFont(FontFactory.COURIER, 10, 
                                                       Font.BOLD, 
                                                       new Color(0, 0, 0));
    private final Font fontLarge = FontFactory.getFont(FontFactory.COURIER, 13, 
                                                       Font.BOLDITALIC, 
                                                       new Color(0, 0, 0));        
	
    /**
     * Starts the report creation for a set of data crow objects.
     * The creation of the report is run in a seperate thread / process.
     * @param objects data crow objects to report on
     * @param reportType HTML or PDF document type
     */
	public TableToFileExporter(DCObject[] objects, int reportType) {
		this.reportType = reportType;
		this.objects = objects;
		this.target = createFile();

		if (target == null) {
			return;
		}
		
		Thread thread = getExportThread();
        thread.setPriority(5);
        thread.start();
	}
	
    /**
     * Open a file browser. With this browser the user can create the
     * report file. Null can be returned if the users chooses to
     * cancel.
     */
	private File createFile() {
        FileFilter filter = reportType == _HTML ? new FileFilter("html") :
        						   				  new FileFilter("pdf");		
        BrowserDialog dialog = new BrowserDialog(Resources.getText(
                				   DCLabels.lblSelectExportFile), filter);
        return dialog.showCreateFileDialog(null, null);
	}
	
    /**
     * Creates and returns the document header
     */
    private HeaderFooter getDocumentHeader() {
        String moduleName = objects[0].getModule().getName();
        
        Chunk chunk = new Chunk(Resources.getText(DCLabels.lblOverview, 
                          moduleName));
        chunk.setFont(fontLarge);
        HeaderFooter header = new HeaderFooter(new Phrase(chunk), false);
        header.setBorder(0);
        header.setBackgroundColor(new Color(250, 245, 240));
        return header;
    }
    
    /**
     * Creates and returns the document footer. Shows the page numbers,
     * etc.
     */    
    private HeaderFooter getDocumentFooter() {
        Chunk chunk1 = new Chunk(Resources.getText(DCLabels.lblPage) + " ");
        chunk1.setFont(fontMedium);
        Chunk chunk2 = new Chunk("");
        chunk2.setFont(fontMedium);
        
        HeaderFooter footer = new HeaderFooter(new Phrase(chunk1), 
                                               new Phrase(chunk2));
        footer.setBorder(0);
        footer.setBackgroundColor(new Color(250, 245, 240));
        return footer;
    }
    
    /**
     * Creates and returns the front page
     */
    private Paragraph getFrontPage() throws Exception {
        Image image = Image.getInstance(
                        new ImageIcon("./icons_system/logo.jpg").getImage(), 
						null, false);
        
        Paragraph paragraph = new Paragraph();

        Chunk chunk = new Chunk("Data Crow Item Report \n");
        chunk.setFont(fontLarge);
        paragraph.add(chunk);
        
        chunk = new Chunk("Created By Robert Jan van der Waals\n\n" +
                          "This Report was created with " +
                          Resources.getText(DCLabels.lblVersion) + "\n" +
                          "Thanks to the authors of iText\n\n");
        chunk.setFont(fontSmall);
        paragraph.add(chunk);
        paragraph.add(image);
        
        chunk = new Chunk("\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n" +
                          "Find Data Crow at http://sourceforge.net/projects/datacrow/ \n" +
                          "or at http://datacrow.sourceforge.net/index.html");
        chunk.setFont(fontSmall);
        
        paragraph.add(chunk);
        
        return paragraph;        
    }	
	
    /**
     * Actual reporting thread. Writes the data from the data crow object 
     * into the report. If a value is hidden or empty, it is not written
     * to the report
     */
	private Thread getExportThread() {
        Thread thread = new Thread(new Runnable() {
            public void run() {
                Font font = FontFactory.getFont(FontFactory.COURIER, 9, 
                                                Font.NORMAL, 
                                                new Color(255, 255, 255));
            	Document document = new Document();
                try {
                    Color color = new Color(240,245,250);
                    
                    document.addCreator("Data Crow");
                    document.addAuthor("Robert-Jan van der Waals");
                    FileOutputStream out = new FileOutputStream(target);

                    if (reportType == _HTML) {
                        HtmlWriter.getInstance(document, out);
                    } else {
                        PdfWriter.getInstance(document, out);
                    }
                    
                    document.open();
                    document.setFooter(getDocumentFooter());
                    document.setHeader(getDocumentHeader());
                    document.add(getFrontPage());
                    document.newPage();
            
                    for (int j = 0; j < objects.length; j++) {
                        Table iTable = new Table(4);
                        iTable.setSpaceBetweenCells(1);
                        iTable.setSpaceInsideCell(1);
                        iTable.setSpacing(1);    
                        iTable.setBorder(0);
                        
                        DCObject o = objects[j];
                        Messages.setMessage(Resources.getText(DCMessages.msgAddingToReport, 
                                                              o.toString()),
                                            Messages._STANDARD);
                        
                        int[] fieldIdx = o.getVisibleFields();
                        Vector vPicCells = new Vector();
                        for (int i = 0; i < fieldIdx.length; i++) {
                        	
                        	String s = o.getValue(fieldIdx[i]).toString();
                        	if (s.length() > 1) {
                        		int fieldType = o.getFieldType(fieldIdx[i]);
	                            if (fieldType == ComponentFactory._PICTUREFIELD) {
	                            	URL url = Utilities.createCachedImage(s);
                                    
                                    if (url != null) {
    	                            	ImageIcon icon = new ImageIcon(url);
    	                            	
    	                            	Image image = Image.getInstance(
    	                            			      icon.getImage(), null, false);
                                        
                                        if (icon.getIconWidth() > 200 && icon.getIconHeight() > 150) {
                                            image.scaleAbsolute(200, 150);
                                        }
                                        
    	                            	Cell cellPicture = new Cell(image);
                                        cellPicture.setBorder(0);
    	                            	cellPicture.setColspan(2);
    	                            	vPicCells.add(cellPicture);
                                    }
	                            } else {
                                    Chunk chunk = new Chunk(o.getField(fieldIdx[i]).getLabel());
                                    chunk.setFont(fontMedium);
                                    Cell cellField = new Cell(chunk);
                            		cellField.setBorder(0);
                            		iTable.addCell(cellField);

                                    chunk = new Chunk(s);
                                    chunk.setFont(fontSmall);
                            		Cell cellValue = new Cell(chunk);
                            		
                                    cellValue.setColspan(3);
                            		cellValue.setBorder(0);
                            		cellValue.setBackgroundColor(color);
                            		iTable.addCell(cellValue);
	                            }
                        	}
                        }

                        for (Enumeration enum = vPicCells.elements(); enum.hasMoreElements(); ) {
                        	iTable.addCell((Cell) enum.nextElement());
                        }
                        
	                    Paragraph cTitle = new Paragraph(o.toString(), font);
	                    Chapter chapter = new Chapter(cTitle, j + 1);
	                    chapter.add(iTable);            
	                    document.add(chapter);
	                    document.newPage();   
	                }
	                document.close();
                } catch (Exception exp) {
                	document.close();
                    new MessageBox(exp.toString(), MessageBox._ERROR);
                    return;
                } 
                new MessageBox(Resources.getText(DCLabels.lblExportHasFinished), 
                               MessageBox._INFORMATION);
            }
        });
        return thread;
	}
}
