/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 Robert-Jan van der Waals                    *
 *                             rwaals@anytimenow.com                          *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.http;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.RandomAccessFile;
import java.net.HttpURLConnection;
import java.net.URL;

import net.sf.dc.core.Repository;
import net.sf.ext.jflexsett.settings.Settings;
import net.sf.ext.util.Base64;

/**
 * HttpURLConnection implementation. Proxy is implemented
 * 
 * @author Robert-Jan van der Waals
 * @since 1.3
 * @version DC1.00
 */
public class HttpConnection {

    /**
     * Read the page from the given location to a String
     * @param url the target
     */
    public static final String getString(URL url) throws IOException {
        HttpURLConnection uc = getConnection(url);
        BufferedInputStream in = new BufferedInputStream(uc.getInputStream());       
        InputStreamReader reader = new InputStreamReader(in);
        
        StringBuffer sb = new StringBuffer();
        int c;
        while ((c = reader.read()) != -1) {
            sb.append((char) c);
        }    
        reader.close();
        in.close();
        uc.disconnect();
        return sb.toString();
    }
    
    /**
     * Read the bytes from the given location to a String.
     * Used for images.
     * @param url the target
     */    
    public static final byte[] getBytes(URL url) throws IOException {
        HttpURLConnection uc = getConnection(url);
        int contentLength = uc.getContentLength();
        
        InputStream raw = uc.getInputStream();
        BufferedInputStream in  = new BufferedInputStream(raw);
        byte[] data = new byte[contentLength];
        int bytesRead = 0;
        int offset = 0;
        while (offset < contentLength) {
            bytesRead = in.read(data, offset, data.length-offset);
            if (bytesRead == -1) {
                break;
            }
            offset += bytesRead;
        }
        in.close();
        
        if (offset != contentLength) {
            throw new IOException("Only read " + offset + " bytes; Expected " + 
                                  contentLength + " bytes");
        }     
        uc.disconnect();
        return data;
    }
    
    /**
     * Retrieves a picture from the net and returns this as a BASE64 string
     * @param url target url
     */
    public static final String getImageAsBase64(URL url) throws IOException {
        byte[] bytes = HttpConnection.getBytes(url);
        String image = String.valueOf(Base64.encode(bytes));
        if (image.length() < 1200) {
            return "";
        } else {
            return image;   
        }
    }
    
    /**
     * Get a Http connection with the correct proxy settings
     * @param url the target url
     */
    private static final HttpURLConnection getConnection(URL url) throws IOException {
        String proxy = Settings.getValueAsString(Repository.settings.stProxyServerName);
        int proxyPort = Settings.getValueAsInt(Repository.settings.stProxyServerPort);
        
        if (proxy.length() > 0 && proxyPort > 0) {
        	System.setProperty("proxySet", "true");
        	System.setProperty("http.proxyHost", proxy);
        	System.setProperty("http.proxyPort", "" + proxyPort);
        } else {
            System.setProperty("proxySet", "false");
            System.setProperty("http.proxyHost", "");
            System.setProperty("http.proxyPort", "" + -1);
        }
        
        HttpURLConnection uc = (HttpURLConnection) url.openConnection();        
        return uc;
    }    
    
    // Used for testing
    public static void main(String[] args) {
        try {
            URL url = new URL("http://images.amazon.com/images/P/B0000A1VER.01.PT01.LZZZZZZZ.jpg");
            byte[] img = HttpConnection.getBytes(url);
            
            RandomAccessFile raf = new RandomAccessFile("test.jpg", "rw");
            raf.write(img);
            raf.close();
            
            url = new URL("http://www.amazon.com/exec/obidos/tg/detail/-/B0000A1VER/002-2890503-3488811");
            String page = HttpConnection.getString(url);
            System.out.print(page);
        } catch (Exception exp) {
            exp.printStackTrace();
        }
    }    
}
