/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 Robert-Jan van der Waals                    *
 *                             rwaals@anytimenow.com                          *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.messages;

import java.util.Date;
import java.util.LinkedList;

import net.sf.dc.core.Repository;
import net.sf.dc.resources.DCMessages;
import net.sf.ext.util.Resources;

/**
 * Message queue. All classes can store a message in this class.
 * The message will be queued and written to the logfile and logform
 * when the program has time to do so. The writing and displaying of
 * messages is done by the class MessageWriter.
 * 
 * Last in last out mechanism is used
 *
 * @author Robert-Jan van der Waals
 * @since 1.3
 * @version DC0.50
 */
public class Messages {

    // Unspecief message type
	public static final int _STANDARD = 0;
	// Dateless message type
	public static final int _LEAVEDATELESS = 1;
	// Only the message text is displayed (indented)
	public static final int _SIMPLE = 2;
    // Query
	public static final int _QUERY = 3;
    // Unindented message type
	public static final int _TABLESS = 4;
    // Error message type
	public static final int _ERROR = 5;

	private static LinkedList lMessagePool = new LinkedList();

    // indicates if there is a new message waiting to be processed
	public static boolean isNew = false;

	private static Thread writeToLog;
    private static String sError =  Resources.getText(DCMessages.msgError);

    /**
     * Starts the message writer thread
     * @see MessagesWriter
     */
	public Messages() {
		writeToLog = new Thread(new MessagesWriter());
		writeToLog.setDaemon(true);
		writeToLog.start();
	}
    
    /**
     * Adds an error message to the message pool
     * @param throwable error to log
     */
    public synchronized static void setMessage(Throwable throwable) {
        StringBuffer sb = new StringBuffer(" ");
        sb.append(throwable.getMessage());
        StackTraceElement[] elements = throwable.getStackTrace();
        for (int i = 0; i < elements.length; i++) {
            sb.append(Repository.operators.opTab);
            sb.append(elements[i].toString());
            sb.append("\n");
        }
        setMessage(sb.toString(), _ERROR);
    }    

    /**
     * Adds a message to the message pool
     * @param message message String
     * @param type message type
     */
	public synchronized static void setMessage(String message, int type) {
        if (message.length() > 500) {
            message = message.substring(0, 255);
        }
        
        StringBuffer sb = new StringBuffer();
		switch (type) {
			case _STANDARD :
                sb.append(Repository.operators.opTab);
                sb.append(getCurrentTime());
                sb.append(Repository.operators.opSeparator);
                sb.append(message);
                sb.append('\n');
				break;
            case _SIMPLE :
			case _LEAVEDATELESS :
                sb.append(Repository.operators.opTab);
                sb.append(message);
                sb.append('\n');
				break;
			case _QUERY :
                sb.append('\n');
                sb.append(message);
                sb.append('\n');
                sb.append('\n');
				break;
			case _TABLESS :
                sb.append(message);
				break;
			case _ERROR :
                sb.append('\n');
                sb.append(Repository.operators.opTab);
                sb.append(Repository.operators.opTab);
                sb.append(Repository.operators.opTab);
                sb.append(sError);
                sb.append(message);
                sb.append('\n');
                sb.append('\n');
				break;
		}
		lMessagePool.add(sb.toString());
        sb = null;
		isNew = true;
	}

    /**
     * Gets a message from the pool. Always gets the first
     * (last in, last out)
     */
	public synchronized static String getNewMessages() {
		Object message = null;
		if (lMessagePool.size() <= 0) {
			isNew = false;
			return null;
		} else {
			message = lMessagePool.removeFirst();
		}
		return message.toString();
	}

    /**
     * Gets the current time to use in the log file
     */
	private static String getCurrentTime() {
        Date date = new Date(System.currentTimeMillis());
		String sCurrentTime = date.toString();
		return sCurrentTime;
	}
}

