/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 Robert-Jan van der Waals                    *
 *                             rwaals@anytimenow.com                          *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.processes;

import java.io.File;
import java.util.Enumeration;
import java.util.LinkedList;
import java.util.Vector;

import net.sf.dc.console.MessageBox;
import net.sf.dc.core.objects.DCObject;
import net.sf.dc.core.objects.DriveReader;
import net.sf.dc.core.objects.Requestor;
import net.sf.dc.core.objects.RequestorCollection;
import net.sf.dc.messages.Messages;
import net.sf.dc.messages.MessagesWriter;
import net.sf.dc.resources.DCMessages;
import net.sf.dc.wf.WorkFlow;
import net.sf.ext.util.Resources;

/**
 * Imports data from a CD (or harddisk) location into the insert table
 * of the owner module
 * 
 * @author Robert-Jan van der Waals
 * @since 1.3
 * @version DC1.00
 */
public abstract class ImportCDProcess extends Thread {

    protected static final WorkFlow wf = new WorkFlow();

    protected String sourceDir = "";
    protected boolean confirmEachDir = false;
    protected String cdName = "";

    /**
     * Initializes (but does  not run) the CD import process
     * @param sourceDir directory which is used as source for the
     *        import
     * @param confirmEachDir indicates whether for each directory
     *        confirmation should be asked if the underlying data
     *        of this directory should be read or not.
     * @param nameOfCD cd name (used as the location name)
     */    
    public ImportCDProcess( String sourceDir,
                            boolean confirmEachDir,
                            String cdName) {
    
        if (!sourceDir.endsWith(File.separator)) {
            this.sourceDir = sourceDir;
        } else {
            this.sourceDir = sourceDir.substring(0, sourceDir.length() -1);
        }
        this.confirmEachDir = confirmEachDir;
        this.cdName = cdName;
    }
    
    /**
     * Starts the CD Import process
     */
    public void run() {
        MessagesWriter._windowShow = true;
        MessagesWriter._windowToFront = true;
        DriveReader cdRom = new DriveReader();
        try {
            Vector filesAndDirs = cdRom.getInformation( sourceDir, 
                                                        confirmEachDir,
                                                        true,
                                                        true,
                                                        null);
            writeOutput(filesAndDirs);
        } catch (Exception exp) {
            exp.printStackTrace();
            MessagesWriter._useLogFile = true;
            Messages.setMessage(exp);
            new MessageBox(Resources.getText(DCMessages.msgProcessCrash), 
                           MessageBox._ERROR);
            finish();
        }
        new MessageBox(Resources.getText(DCMessages.msgImportCompleted), 
                       MessageBox._INFORMATION);
        finish();
    }    
    
    /**
     * Parses the source files for information and adds the parsed
     * information to the insert table of the owner module.
     * @param filenames files to be parsed
     */
    public void writeOutput(Vector filenames) {
        try {
            new MessageBox( Resources.getText(DCMessages.msgCreateOutput), 
                            MessageBox._INFORMATION); 
            for (Enumeration enum = filenames.elements(); enum.hasMoreElements(); ) {
                String sFilename = (String) enum.nextElement();
                DCObject o = parseInformation(sFilename);
                
                if (o != null) {
                	LinkedList values = new LinkedList();
                	values.add(o);
                    
                	Requestor req = new Requestor(getUITargetClass(),
                								  "addRow",
                                                  null);
                	RequestorCollection reqCol = new RequestorCollection();
                	reqCol.addRequestor(req);
                	wf.displayResults(values, reqCol, true);
                }
            }
        } catch (Exception exp) {
            MessagesWriter._useLogFile = true;
            new MessageBox(Resources.getText(DCMessages.msgProcessCrash), 
                           MessageBox._ERROR);
            Messages.setMessage(exp);
            filenames.clear();
            filenames = null;
            finish();
        }
        filenames.clear();
        filenames = null;
    }    
    
    /**
     * Parses the file for its information. Needs to be implemented
     * by the extending classes 
     * @param sFilename file to be parsed
     */
    protected abstract DCObject parseInformation(String sFilename);
    
    /**
     * The UI class which will be updated with the results of the 
     * import 
     */
    protected abstract Class getUITargetClass();

    /**
     * Stops the thread
     */
    protected void finish() {
        MessagesWriter._useLogFile = true;
        Messages.setMessage(
            Resources.getText(DCMessages.msgImportStops), 
            Messages._STANDARD);
    }
}
