/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 Robert-Jan van der Waals                    *
 *                             rwaals@anytimenow.com                          *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.processes;

import java.io.File;

import net.sf.dc.console.music.InsertMusicPanel;
import net.sf.dc.core.objects.DCObject;
import net.sf.dc.core.objects.music.Music;
import net.sf.dc.messages.Messages;
import net.sf.dc.resources.DCMessages;
import net.sf.ext.entagged.libraries.audioformats.AudioFormatManager;
import net.sf.ext.entagged.libraries.audioformats.EncodingInfo;
import net.sf.ext.entagged.libraries.audioformats.interfaces.EncodingInfoReader;
import net.sf.ext.entagged.libraries.audioformats.interfaces.Tag;
import net.sf.ext.entagged.libraries.audioformats.interfaces.TagReaderInterface;
import net.sf.ext.helliker.id3.MP3File;
import net.sf.ext.util.Resources;

/**
 * Imports data from a CD (or harddisk) location into the insert table
 * of the owner module
 * @author Robert-Jan van der Waals
 * @since 1.3
 * @version DC0.50
 */
public class ImportMusicCD extends ImportCDProcess {

    /**
     * Initializes (but does  not run) the Movie CD import process
     * @param sourceDir directory which is used as source for the
     *        import
     * @param confirmEachDir indicates whether for each directory
     *        confirmation should be asked if the underlying data
     *        of this directory should be read or not.
     * @param nameOfCD cd name (used as the location name)
     */   
    public ImportMusicCD( String sourceDir,
            			  boolean confirmEachDir,
						  String cdName) {
    	super(sourceDir, confirmEachDir, cdName);
    }

    /**
     * The UI class which will be updated with the results of the 
     * import 
     */    
    protected Class getUITargetClass() {
    	return InsertMusicPanel.class;
    }

    /**
     * Parses the source file for information. 
     */   
    protected DCObject parseInformation(String sFile) {
    	if (!sFile.toLowerCase().endsWith("mp3") && 
    	    !sFile.toLowerCase().endsWith("ogg") &&
    	    !sFile.toLowerCase().endsWith("flc") &&
    	    !sFile.toLowerCase().endsWith("wav") &&
			!sFile.toLowerCase().endsWith("flac")) {
    		return null;
    	}
    	
        Music music = new Music();
        music.setValue(Music._K_LOCATION, cdName);

        String sFilename = sFile.substring(2, sFile.length());
        music.setValue(Music._A_FILENAME, sFilename);
        
        try {
            File file = new File(sFile);
            EncodingInfo encInfo = null;
            try {
                EncodingInfoReader reader = AudioFormatManager.getEncodingInfoReader(file);
                encInfo = reader.read(file);
            } catch (Exception exp) {
                return getMp3Information(file, music);
            }

            music.setValue(Music._H_QUALITY, new Integer(encInfo.getBitrate()));
            music.setValue(Music._G_PLAYLENGTH, new Integer(encInfo.getTimeLength()));
            music.setValue(Music._I_ENCODING, encInfo.getEncodingType());

            TagReaderInterface[] tagReaders = AudioFormatManager.getTagReaders(file);
            for (int i = 0; i < tagReaders.length; i++) {
                Tag tag = tagReaders[0].read(file);
                if (tag != null) {
                    music.setValue(Music._B_TITLE, tag.getTitle());
                    music.setValue(Music._C_ARTIST, tag.getArtist());
                    music.setValue(Music._D_ALBUM, tag.getAlbum());
                    music.setValue(Music._E_YEAR, tag.getYear());
                    music.setValue(Music._F_GENRE, tag.getGenre());                    
                }
            }
            return music;
        } catch (Exception exp) {
        	Messages.setMessage(Resources.getText(DCMessages.msgDidNotParseFileBecause, sFile),
				    Messages._ERROR);
        	return null;
        }
    }

    /**
     * Parses an Mp3 file for information
     * @param file the file to be parsed
     * @param music the Music object to which the data shoudl be added
     */
    private Music getMp3Information(File file, Music music) throws Exception {
        MP3File mp3File = new MP3File(file);
        
        music.setValue(Music._H_QUALITY, new Integer(mp3File.getBitRate()));
        music.setValue(Music._G_PLAYLENGTH, new Integer(new Long(mp3File.getPlayingTime()).intValue()));        
        music.setValue(Music._B_TITLE, mp3File.getTitle());
        music.setValue(Music._C_ARTIST, mp3File.getArtist());
        music.setValue(Music._D_ALBUM, mp3File.getAlbum());
        music.setValue(Music._E_YEAR, mp3File.getYear());
        music.setValue(Music._F_GENRE, mp3File.getGenre());
        music.setValue(Music._I_ENCODING, mp3File.getMPEGLayer() + " " + mp3File.getMPEGVersion());  
        return music;   
    }
}
