/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 Robert-Jan van der Waals                    *
 *                             rwaals@anytimenow.com                          *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.wf;

import java.sql.ResultSet;
import java.sql.Types;
import java.util.LinkedList;

import javax.swing.SwingUtilities;

import net.sf.dc.core.DataCrow;
import net.sf.dc.core.objects.DCCollection;
import net.sf.dc.core.objects.DCObject;
import net.sf.dc.core.objects.RequestorCollection;
import net.sf.dc.core.objects.movie.Movie;
import net.sf.dc.core.objects.music.Music;
import net.sf.dc.core.objects.software.Software;
import net.sf.dc.core.objects.software.SoftwareCategory;
import net.sf.dc.db.Database;
import net.sf.dc.db.Query;
import net.sf.dc.db.QueryOptions;
import net.sf.dc.messages.Messages;
import net.sf.ext.util.Base64;

/**
 * This is the layer between the UI and the database. Checks for data
 * integrity and checks on bussiness rules are done here
 *
 * @author Robert-Jan van der Waals
 * @since 1.3
 * @version DC0.50
 */
public class WorkFlow {

    private static Database db;

    /**
     * Initializes the database
     */
    public static void initDatabase() {
        if (db == null) {
            db = new Database();
        }
    }

    /**
     * Updates a collection of data crow objects
     * @param collection data crow object collection
     * @param requestors requests to be executed after update
     */  
    public void updateCollection(DCCollection collection,
                                 RequestorCollection requestors) {
        initDatabase();
        db.updateCollection(collection, requestors);
    }

    /**
     * Inserts a collection of data crow objects into the database
     * @param collection data crow object collection
     * @param requestors requests to be executed after insert
     */
    public void insertCollection(DCCollection collection,
                                 RequestorCollection requestors) {
        initDatabase();
        db.insertCollection(collection, requestors);
    }

    /**
     * Deletes a collection of data crow objects from the database
     * @param collection data crow object collection
     * @param requestors requests to be executed after delete
     */  
    public void deleteCollection(DCCollection collection,
                                 RequestorCollection requestors) {
        initDatabase();
        db.deleteCollection(collection, requestors);
    }

    /**
     * Runs a predefined query
     *
     * @param sQuery query statement
     * @param requestors  requests to the UI
     */
    public void runQuery(String sQuery, RequestorCollection requestors) {
        initDatabase();
        db.runQuery(sQuery, requestors);
    }

    /**
     * Searches for values. The query is put in the query queue. The results are
     * past to the caller by using the RequestorCollection
     *
     * @param o the object which is used as the condition
     * @param options query options
     * @param requestors requestors for the result of the query (UI)
     */
    public void searchValues(DCObject o, QueryOptions queryOptions, RequestorCollection requestors) {
        initDatabase();
        db.searchValues(o, queryOptions, requestors);
    }

    /**
     * Inserts new values. The query is put in the query queue. Actions are
     * past to the caller by using the RequestorCollection
     *
     * @param o the object to be inserted
     * @param requestors    requestors which depend on the outcome of the query
     */
    public void insertValues(DCObject o, RequestorCollection requestors) {
        initDatabase();
        db.insertValues(o, requestors);
    }

    /**
     * Updates values. The query is put in the query queue. Actions are
     * past to the caller by using the RequestorCollection
     *
     * @param o  the object to be updated
     * @param requestors requestors which depend on the outcome of the query
     */
    public void updateValues(DCObject o, RequestorCollection requestors) {
        initDatabase();
        db.updateValues(o, requestors);
    }

    /**
     * Deletes teh object. The query is put in the query queue. Actions are
     * past to the caller by using the RequestorCollection
     *
     * @param o the object to be updated
     * @param requestors requestors which depend on the outcome of the query
     */    
    public void deleteValues(DCObject o, RequestorCollection requestors) {
        initDatabase();
        db.deleteValues(o, requestors);
    }

    /**
     * Runs a query directly, bypassing the query queue
     * @param o Data Crow object
     */
    public LinkedList runQueryDirect(DCObject o) {
        initDatabase();
        return db.runQueryDirect(o);
    }

    /**
     * Does not put the query in a queue but executes it immediately.
     * Errors are returned to the caller (not thrown)
     * @param Query Query object
     */
    public LinkedList runQueryDirect(Query query) {
        initDatabase();
        return db.runQueryDirect(query);
    }

    /**
     * Does not put the query in a queue but executes it immediately.
     * Does not close the resultset!
     * @param sQuery query statement
     */
    public ResultSet runQueryDirectUnclosed(String sQuery) throws Exception {
        initDatabase();
        return db.runQueryDirectUnclosed(sQuery);
    }

    /**
     * Does not put the query in a queue but executes it immediately.
     * @param sQuery query statement
     */
    public LinkedList runQueryDirect(String sQuery) {
        initDatabase();
        return db.runQueryDirect(sQuery);
    }

    /**
     * Passes requests thru to the UI (<url>MainFrame</url>)
     * @param vTable data to be shown
     * @param requestorObjects collection of requests
     */
    public void displayResults(LinkedList data, RequestorCollection requestorObjects, boolean qryWasSuccess) {
        Runnable runMe = new UIUpdater(data, requestorObjects, qryWasSuccess);
        SwingUtilities.invokeLater(runMe);
        runMe = null;
    }
    
    private class UIUpdater implements Runnable {
        private final LinkedList values;
        private final RequestorCollection requestors;
        private final boolean qryWasSuccess;
        
        public UIUpdater(LinkedList values, 
                         RequestorCollection requestors,
                         boolean qryWasSuccess) {
        	this.qryWasSuccess = qryWasSuccess;
            this.values = values;
            this.requestors = requestors;
        }
        
        public void run() {
            if (DataCrow.mainFrame == null) {
                return;
            } else {
                DataCrow.mainFrame.displayResults(requestors, 
                                                  values, 
                                                  qryWasSuccess);
            }
        }
    }

    /**
     * Checks if the object is unique
     * @param o object to be checked
     * @param bUpdate specifies if the object exists already in the
     *        database or not
     */
    public boolean checkUniqueness(DCObject o, boolean bUpdate) {
        initDatabase();
        return db.uniqueValues(o, bUpdate);
    }

    /**
     * Converts a resultset to a DataCrowObject Vector
     * @param result result set from a query
     *
     */
    public static LinkedList createDataCrowObjectArray(ResultSet result) {
        LinkedList values = new LinkedList();

        try {
        	result.isLast();
        } catch (Exception exp) {
            return values;
        }
        
        try {
            String sTableName = result.getMetaData().getTableName(1);
            int columns = result.getMetaData().getColumnCount();
            String[] sColumns = new String[columns];

            int k = 0;
            for (int i = 1; i < columns + 1; i++) {
                sColumns[k] = result.getMetaData().getColumnName(i);
                k++;
            }

            while (result.next()) {
                Object[] rowData = new String[columns];
                for (int j = 0; j < sColumns.length; j++) {
                    int type = result.getMetaData().getColumnType(j + 1);
                    Object value = null;
                    
                    if (type == Types.BINARY) {
                        value = result.getBytes(sColumns[j]);
                        if (value != null) {
                            value = Base64.encode((byte[]) value);
                        } 
                    } else {
                        value = result.getString(sColumns[j]);    
                    }
                    rowData[j] = value == null ? "" : value;
                }

                DCObject dcObj = null;
                if (sTableName.toUpperCase().equals("SOFTWARE")) {
                    dcObj = new Software();
                } else if (sTableName.toUpperCase().equals("SOFTWARETYPE")) {
                    dcObj = new SoftwareCategory();
                } else if (sTableName.toUpperCase().equals("MOVIE")) {
                    dcObj = new Movie();
                } else if (sTableName.toUpperCase().equals("MP3")) {
                    dcObj = new Music();
                }
                
                dcObj.setValues(rowData);
                values.add(dcObj);
                rowData = null;
                dcObj = null;
            }
        } catch (Exception exp) {
            Messages.setMessage(exp);
        } 
        return values;
    }
}
