/*
 *  ********************************************************************   **
 *  Copyright notice                                                       **
 *  **																	   **
 *  (c) 2003 Entagged Developpement Team				                   **
 *  http://www.sourceforge.net/projects/entagged                           **
 *  **																	   **
 *  All rights reserved                                                    **
 *  **																	   **
 *  This script is part of the Entagged project. The Entagged 			   **
 *  project is free software; you can redistribute it and/or modify        **
 *  it under the terms of the GNU General Public License as published by   **
 *  the Free Software Foundation; either version 2 of the License, or      **
 *  (at your option) any later version.                                    **
 *  **																	   **
 *  The GNU General Public License can be found at                         **
 *  http://www.gnu.org/copyleft/gpl.html.                                  **
 *  **																	   **
 *  This copyright notice MUST APPEAR in all copies of the file!           **
 *  ********************************************************************
 */
package net.sf.ext.entagged.plugins.audioformats.flac;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.io.UnsupportedEncodingException;
import java.nio.MappedByteBuffer;
import java.nio.channels.FileChannel;
import java.util.Hashtable;

import net.sf.ext.entagged.libraries.audioformats.interfaces.Tag;
import net.sf.ext.entagged.libraries.audioformats.interfaces.TagReaderInterface;
import net.sf.ext.entagged.libraries.exceptions.CannotReadException;
import net.sf.ext.entagged.plugins.audioformats.flac.inforeader.MetadataBlockHeader;


public class FLACReader implements TagReaderInterface {

	private MappedByteBuffer mbb;
	private FLACTag tag = null;

	public Tag read (File f) throws CannotReadException{
		try {
			RandomAccessFile raf = new RandomAccessFile( f, "r" );

			mbb = raf.getChannel().map( FileChannel.MapMode.READ_ONLY, 0, raf.getChannel().size() );
			raf.close();
		} catch ( IOException e ) {
			throw new CannotReadException("Error: Could not access the file -"+f+"- :\n"+e);
		}
		
		tag = null;
		
		if (containsFLACTag( f )) {
            readFLAC( f );
        }
			
		return tag;
	}  

	public String toString() {
		return "FLAC Tag Reader";
	}

	private boolean containsFLACTag( File f ) {
		if ( f.length()==0 ) {
			return false;
		}
        
		mbb.position( 0 );

		byte[] b = new byte[4];
		mbb.get(b);
		String flac = new String(b);
		return flac.equals("fLaC");
	}

	private void readFLAC(File f) {
		byte[] b;
		boolean isLastBlock = false;
		
		while(!isLastBlock) {
			b = new byte[4];
			mbb.get(b);
			MetadataBlockHeader mbh = new MetadataBlockHeader(b);
		
			switch(mbh.getBlockType()) {
				case MetadataBlockHeader.VORBIS_COMMENT : handleVorbisComment(mbh); break;
				default : skipBlock(mbh); break;
			}
			isLastBlock = mbh.isLastBlock();
		}
	}
	
	private void skipBlock(MetadataBlockHeader mbh) {
		mbb.position(mbb.position()+mbh.getDataLength());
	}
	
	private void handleVorbisComment(MetadataBlockHeader mbh) {
		byte[] b = new byte[mbh.getDataLength()];
		mbb.get(b);
		
		int pos = 0;
		int vendorStringLength = u(b[0]) + (u(b[1]) << 8 ) + ( u( b[2] ) << 16 ) + ( u( b[3] ) << 24 );
		pos += 4;
		
		String vendorString = new String( b, 4,vendorStringLength );
		pos += vendorStringLength;
	
		int userComments = u(b[pos]) + (u(b[pos+1]) << 8 ) + ( u( b[pos+2] ) << 16 ) + ( u( b[pos+3] ) << 24 );
		pos += 4;
		
		Hashtable ht = new Hashtable(10);
	
		for ( int i = 0; i < userComments; i++ ) {
			int commentLength = u(b[pos]) + (u(b[pos+1]) << 8 ) + ( u( b[pos+2] ) << 16 ) + ( u( b[pos+3] ) << 24 );
			pos += 4;
			
			// Attempt to convert the comment bytes to UTF-8 - if that fails,
			// just use the platforms default charset.
			String comment = null;
			try {
				comment = new String( b, pos, commentLength, "UTF-8" );
			} catch ( UnsupportedEncodingException e ) {
				comment = new String( b, pos, commentLength );
			}
			pos += commentLength;
			
			String[] splitComment = comment.split( "=" );
			if(splitComment.length>1)
				ht.put(splitComment[0],splitComment[1]);
		}
		
		tag = new FLACTag();
	
		tag.setTitle(  (ht.get("TITLE")==null) ? "" : (String)ht.get("TITLE") );
		tag.setArtist((ht.get("ARTIST")==null) ? "" : (String)ht.get("ARTIST"));
		tag.setAlbum((ht.get("ALBUM")==null) ? "" : (String)ht.get("ALBUM"));
		tag.setYear((ht.get("DATE")==null) ? "" : (String)ht.get("DATE"));
		
		tag.setComment(
			(ht.get("COMMENT")==null) ? 
				(
				(ht.get("DESCRIPTION")==null) ? "" : (String)ht.get("DESCRIPTION")
				)
				: (String)ht.get("COMMENT"));
		
		tag.setTrackNumber(
			(ht.get("TRACK")==null) ? 
				(
				(ht.get("TRACKNUMBER")==null) ? "" : (String)ht.get("TRACKNUMBER")
				)
				: (String)ht.get("TRACK"));
		
		tag.setGenre((ht.get("GENRE")==null) ? "" : (String)ht.get("GENRE"));
		
		tag.setVendor(vendorString);
	}
	
	private int u(int i) {
		return i & 0xFF;
	}
}

